import re
import json
import ftfy
import html
import itertools
from functools import lru_cache
from math import log10
from unidecode import unidecode
from .texts._tatabahasa import rules_normalizer, hujung
from .texts._regex import _expressions, _money
from .texts._english_words import _english_words
from ._utils._paths import PATH_PREPROCESSING, S3_PATH_PREPROCESSING
from ._utils._utils import check_file, check_available
from .stem import naive

_annotate = [
    'hashtag',
    'allcaps',
    'elongated',
    'repeated',
    'emphasis',
    'censored',
]

_normalize = list(_expressions.keys())

REGEX_TOKEN = re.compile(r'\b[a-z]{2,}\b')
NGRAM_SEP = '_'


def get_normalize():
    return _normalize


def get_annotate():
    return _annotate


def _case_of(text):
    return (
        str.upper
        if text.isupper()
        else str.lower
        if text.islower()
        else str.title
        if text.istitle()
        else str
    )


def _naive_stem(word):
    hujung_result = [e for e in hujung if word.endswith(e)]
    if len(hujung_result):
        hujung_result = max(hujung_result, key = len)
        if len(hujung_result):
            word = word[: -len(hujung_result)]
    return word


def unpack_english_contractions(text):
    """
    Replace *English* contractions in ``text`` str with their unshortened forms.
    N.B. The "'d" and "'s" forms are ambiguous (had/would, is/has/possessive),
    so are left as-is.
    Important Note: The function is taken from textacy (https://github.com/chartbeat-labs/textacy).
    """

    text = re.sub(
        r"(\b)([Aa]re|[Cc]ould|[Dd]id|[Dd]oes|[Dd]o|[Hh]ad|[Hh]as|[Hh]ave|[Ii]s|[Mm]ight|[Mm]ust|[Ss]hould|[Ww]ere|[Ww]ould)n't",
        r'\1\2 not',
        text,
    )
    text = re.sub(
        r"(\b)([Hh]e|[Ii]|[Ss]he|[Tt]hey|[Ww]e|[Ww]hat|[Ww]ho|[Yy]ou)'ll",
        r'\1\2 will',
        text,
    )
    text = re.sub(
        r"(\b)([Tt]hey|[Ww]e|[Ww]hat|[Ww]ho|[Yy]ou)'re", r'\1\2 are', text
    )
    text = re.sub(
        r"(\b)([Ii]|[Ss]hould|[Tt]hey|[Ww]e|[Ww]hat|[Ww]ho|[Ww]ould|[Yy]ou)'ve",
        r'\1\2 have',
        text,
    )
    text = re.sub(r"(\b)([Cc]a)n't", r'\1\2n not', text)
    text = re.sub(r"(\b)([Ii])'m", r'\1\2 am', text)
    text = re.sub(r"(\b)([Ll]et)'s", r'\1\2 us', text)
    text = re.sub(r"(\b)([Ww])on't", r'\1\2ill not', text)
    text = re.sub(r"(\b)([Ss])han't", r'\1\2hall not', text)
    text = re.sub(r"(\b)([Yy])(?:'all|a'll)", r'\1\2ou all', text)
    return text


def _get_expression_dict():
    return {
        k.lower(): re.compile(_expressions[k]) for k, v in _expressions.items()
    }


class SocialTokenizer:
    def __init__(self, lowercase = False, **kwargs):
        """
        Args:
            lowercase (bool): set to True in order to lowercase the text
        Kwargs ():
            emojis (bool): True to keep emojis
            urls (bool): True to keep urls
            tags (bool): True to keep tags: <tag>
            emails (bool): True to keep emails
            users (bool): True to keep users handles: @cbaziotis
            hashtags (bool): True to keep hashtags
            cashtags (bool): True to keep cashtags
            phones (bool): True to keep phones
            percents (bool): True to keep percents
            money (bool): True to keep money expressions
            date (bool): True to keep date expressions
            time (bool): True to keep time expressions
            acronyms (bool): True to keep acronyms
            emoticons (bool): True to keep emoticons
            censored (bool): True to keep censored words: f**k
            emphasis (bool): True to keep words with emphasis: *very* good
            numbers (bool): True to keep numbers
        """

        self.lowercase = lowercase
        pipeline = []
        self.regexes = _expressions

        emojis = kwargs.get('emojis', True)
        urls = kwargs.get('urls', True)
        tags = kwargs.get('tags', True)
        emails = kwargs.get('emails', True)
        users = kwargs.get('users', True)
        hashtags = kwargs.get('hashtags', True)
        cashtags = kwargs.get('cashtags', True)
        phones = kwargs.get('phones', True)
        percents = kwargs.get('percents', True)
        money = kwargs.get('money', True)
        date = kwargs.get('date', True)
        time = kwargs.get('time', True)
        acronyms = kwargs.get('acronyms', True)
        emoticons = kwargs.get('emoticons', True)
        censored = kwargs.get('censored', True)
        emphasis = kwargs.get('emphasis', True)
        numbers = kwargs.get('numbers', True)

        if urls:
            pipeline.append(self.regexes['url'])

        if tags:
            pipeline.append(self.regexes['tag'])

        if emails:
            pipeline.append(self.wrap_non_matching(self.regexes['email']))

        if users:
            pipeline.append(self.wrap_non_matching(self.regexes['user']))

        if hashtags:
            pipeline.append(self.wrap_non_matching(self.regexes['hashtag']))

        if cashtags:
            pipeline.append(self.wrap_non_matching(self.regexes['cashtag']))

        if phones:
            pipeline.append(self.wrap_non_matching(self.regexes['phone']))

        if percents:
            pipeline.append(self.wrap_non_matching(self.regexes['percent']))

        if money:
            pipeline.append(self.wrap_non_matching(self.regexes['money']))

        if date:
            pipeline.append(self.wrap_non_matching(self.regexes['date']))

        if time:
            pipeline.append(self.wrap_non_matching(self.regexes['time']))

        if acronyms:
            pipeline.append(self.wrap_non_matching(self.regexes['acronym']))

        if emoticons:
            pipeline.append(self.regexes['ltr_face'])
            pipeline.append(self.regexes['rtl_face'])

        if censored:
            pipeline.append(self.wrap_non_matching(self.regexes['censored']))

        if emphasis:
            pipeline.append(self.wrap_non_matching(self.regexes['emphasis']))

        if emoticons:
            pipeline.append(
                self.wrap_non_matching(self.regexes['rest_emoticons'])
            )

        if numbers:
            pipeline.append(self.regexes['number'])

        if emojis:
            pipeline.append(self.regexes['emoji'])

        # any other word
        pipeline.append(self.regexes['word'])

        if emoticons:
            pipeline.append(
                self.wrap_non_matching(self.regexes['eastern_emoticons'])
            )

        # keep repeated puncts as one term
        # pipeline.append(r"")

        pipeline.append('(?:\S)')  # CATCH ALL remaining terms

        self.tok = re.compile(r'({})'.format('|'.join(pipeline)))

    @staticmethod
    def wrap_non_matching(exp):
        return '(?:{})'.format(exp)

    def tokenize(self, text):
        escaped = html.unescape(text)
        tokenized = self.tok.findall(escaped)

        if self.lowercase:
            tokenized = [t.lower() for t in tokenized]

        return tokenized


def _read_stats(gram = 1):
    with open(PATH_PREPROCESSING[gram]['model']) as fopen:
        return json.load(fopen)


class _Pdist(dict):
    @staticmethod
    def default_unk_func(key, total):
        return 1.0 / total

    def __init__(self, data = None, total = None, unk_func = None, **kwargs):
        super().__init__(**kwargs)

        data = data or {}
        for key, count in data.items():
            self[key] = self.get(key, 0) + int(count)

        self.total = float(total or sum(self.values()))
        self.unk_prob = unk_func or self.default_unk_func

    def __call__(self, key):
        if key in self:
            return self[key] / self.total
        else:
            return self.unk_prob(key, self.total)


class _Segmenter:
    def __init__(self, max_split_length = 20):
        self.unigrams = _read_stats(1)
        self.bigrams = _read_stats(2)
        self.N = sum(self.unigrams.values())
        self.L = max_split_length

        self.Pw = _Pdist(self.unigrams, self.N, self.unk_probability)
        self.P2w = _Pdist(self.bigrams, self.N)

        self.case_split = _get_expression_dict()['camel_split']

    def condProbWord(self, word, prev):
        try:
            return self.P2w[prev + NGRAM_SEP + word] / float(self.Pw[prev])
        except KeyError:
            return self.Pw(word)

    @staticmethod
    def unk_probability(key, total):
        return 10.0 / (total * 10 ** len(key))

    @staticmethod
    def combine(first, rem):
        (first_prob, first_word) = first
        (rem_prob, rem_words) = rem
        return first_prob + rem_prob, [first_word] + rem_words

    def splits(self, text):
        return [
            (text[: i + 1], text[i + 1 :])
            for i in range(min(len(text), self.L))
        ]

    @lru_cache(maxsize = 65536)
    def find_segment(self, text, prev = '<S>'):
        if not text:
            return 0.0, []
        candidates = [
            self.combine(
                (log10(self.condProbWord(first, prev)), first),
                self.find_segment(rem, first),
            )
            for first, rem in self.splits(text)
        ]
        return max(candidates)

    @lru_cache(maxsize = 65536)
    def segment(self, word):
        if word.islower():
            return ' '.join(self.find_segment(word)[1])
        else:
            return self.case_split.sub(r' \1', word).lower()


class _Preprocessing:
    def __init__(
        self,
        normalize = [
            'url',
            'email',
            'percent',
            'money',
            'phone',
            'user',
            'time',
            'date',
            'number',
        ],
        annotate = [
            'allcaps',
            'elongated',
            'repeated',
            'emphasis',
            'censored',
            'hashtag',
        ],
        lowercase = True,
        fix_unidecode = True,
        expand_hashtags = True,
        expand_english_contractions = True,
        remove_postfix = True,
        maxlen_segmenter = 20,
        translator = None,
        speller = None,
    ):
        self._fix_unidecode = fix_unidecode
        self._normalize = normalize
        self._annotate = annotate
        self._remove_postfix = remove_postfix
        self._regexes = _get_expression_dict()
        self._expand_hashtags = expand_hashtags
        self._tokenizer = SocialTokenizer(lowercase = lowercase).tokenize
        if self._expand_hashtags:
            self._segmenter = _Segmenter(maxlen_segmenter)
        self._expand_contractions = expand_english_contractions
        self._all_caps_tag = 'wrap'
        self._translator = translator
        self._speller = speller

    def _add_special_tag(self, m, tag, mode = 'single'):

        if isinstance(m, str):
            text = m
        else:
            text = m.group()

        if mode == 'single':
            return ' {} <{}> '.format(text, tag)
        elif mode == 'wrap':
            return ' '.join([' <{}> {} </{}> '.format(tag, text, tag)]) + ' '
        elif mode == 'every':
            tokens = text.split()
            processed = ' '.join([' {} <{}> '.format(t, tag) for t in tokens])
            return ' ' + processed + ' '

    @lru_cache(maxsize = 65536)
    def _handle_hashtag_match(self, m):
        expanded = m.group()[1:]
        if self._expand_hashtags:
            if expanded.islower():
                expanded = self._segmenter.segment(expanded)
                expanded = ' '.join(expanded.split('-'))
                expanded = ' '.join(expanded.split('_'))

            else:
                expanded = self._regexes['camel_split'].sub(r' \1', expanded)
                expanded = expanded.replace('-', '')
                expanded = expanded.replace('_', '')

        if 'hashtag' in self._annotate:
            expanded = self._add_special_tag(expanded, 'hashtag', mode = 'wrap')

        return expanded

    @lru_cache(maxsize = 65536)
    def _handle_repeated_puncts(self, m):
        text = m.group()
        text = ''.join(sorted(set(text), reverse = True))

        if 'repeated' in self._annotate:
            text = self._add_special_tag(text, 'repeated')

        return text

    @lru_cache(maxsize = 65536)
    def _handle_generic_match(self, m, tag, mode = 'every'):
        text = m.group()
        text = self._add_special_tag(text, tag, mode = mode)

        return text

    def _handle_elongated_match(self, m):
        text = m.group()
        text = self._regexes['normalize_elong'].sub(r'\1\1', text)
        if self._speller and text.lower() not in _english_words:
            if hasattr(self._speller, 'normalize_elongated'):
                text = _case_of(text)(
                    self._speller.normalize_elongated(text.lower())
                )
            else:
                text = _case_of(text)(self._speller.correct(text.lower()))
        if 'elongated' in self._annotate:
            text = self._add_special_tag(text, 'elongated')
        return text

    @lru_cache(maxsize = 65536)
    def _handle_emphasis_match(self, m):
        text = m.group().replace('*', '')
        if 'emphasis' in self._annotate:
            text = self._add_special_tag(text, 'emphasis')

        return text

    def _dict_replace(self, wordlist, _dict):
        return [_dict.get(w, w) for w in wordlist]

    @staticmethod
    def text(wordlist):
        in_hashtag = False
        _words = []
        for word in wordlist:

            if word == '<hashtag>':
                in_hashtag = True
            elif word == '</hashtag>':
                in_hashtag = False
            elif word in {'<allcaps>', '</allcaps>'} and in_hashtag:
                continue

            _words.append(word)

        return _words

    def process(self, text):
        text = re.sub(r' +', ' ', text)
        if self._fix_unidecode:
            text = ftfy.fix_text(text)

        for item in self._normalize:
            text = self._regexes[item].sub(
                lambda m: ' ' + '<' + item + '>' + ' ', text
            )

        text = self._regexes['hashtag'].sub(
            lambda w: self._handle_hashtag_match(w), text
        )

        if 'allcaps' in self._annotate:
            text = self._regexes['allcaps'].sub(
                lambda w: self._handle_generic_match(
                    w, 'allcaps', mode = self._all_caps_tag
                ),
                text,
            )
        if 'elongated' in self._annotate:
            text = self._regexes['elongated'].sub(
                lambda w: self._handle_elongated_match(w), text
            )
        if 'repeated' in self._annotate:
            text = self._regexes['repeat_puncts'].sub(
                lambda w: self._handle_repeated_puncts(w), text
            )
        if 'emphasis' in self._annotate:
            text = self._regexes['emphasis'].sub(
                lambda w: self._handle_emphasis_match(w), text
            )
        if 'censored' in self._annotate:
            text = self._regexes['censored'].sub(
                lambda w: self._handle_generic_match(w, 'censored'), text
            )
        if self._expand_contractions:
            text = unpack_english_contractions(text)

        text = re.sub(r' +', ' ', text)
        text = self.text(text.split())
        text = ' '.join(text)
        text = self._tokenizer(text)
        text = self._dict_replace(text, rules_normalizer)
        if self._translator:
            text = self._dict_replace(text, self._translator)
        if self._remove_postfix:
            text = [
                _naive_stem(w) if w not in _english_words else w for w in text
            ]

        return text


def preprocessing(
    normalize = [
        'url',
        'email',
        'percent',
        'money',
        'phone',
        'user',
        'time',
        'date',
        'number',
    ],
    annotate = [
        'allcaps',
        'elongated',
        'repeated',
        'emphasis',
        'censored',
        'hashtag',
    ],
    lowercase = True,
    fix_unidecode = True,
    expand_hashtags = True,
    expand_english_contractions = True,
    translate_english_to_bm = True,
    remove_postfix = True,
    maxlen_segmenter = 20,
    validate = True,
    speller = None,
):
    """
    Load Preprocessing class.

    Parameters
    ----------
    normalize: list
        normalizing tokens, can check all supported normalizing at malaya.preprocessing.get_normalize()
    annotate: list
        annonate tokens <open></open>, only accept ['hashtag', 'allcaps', 'elongated', 'repeated', 'emphasis', 'censored']
    lowercase: bool
    fix_unidecode: bool
    expand_hashtags: bool
        expand hashtags using Viterbi algorithm, #mondayblues == monday blues
    expand_english_contractions: bool
        expand english contractions
    translate_english_to_bm: bool
        translate english words to bahasa malaysia words
    remove_postfix: bool
        remove postfix from a word, faster way to get root word
    speller: object
        spelling correction object, need to have a method `correct`
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.


    Returns
    -------
    _Preprocessing : malaya.preprocessing._Preprocessing class
    """

    if any([e not in _normalize for e in normalize]):
        raise ValueError(
            'normalize element not able to recognize, supported normalization can check at get_normalize()'
        )
    if any([e not in _annotate for e in annotate]):
        raise ValueError(
            "annotate only accept ['hashtag', 'allcaps', 'elongated', 'repeated', 'emphasis', 'censored']"
        )
    if not isinstance(lowercase, bool):
        raise ValueError('lowercase must be a boolean')
    if not isinstance(fix_unidecode, bool):
        raise ValueError('fix_unidecode must be a boolean')
    if not isinstance(expand_hashtags, bool):
        raise ValueError('expand_hashtags must be a boolean')
    if not isinstance(expand_english_contractions, bool):
        raise ValueError('expand_english_contractions must be a boolean')
    if not isinstance(translate_english_to_bm, bool):
        raise ValueError('translate_english_to_bm must be a boolean')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')
    if not isinstance(remove_postfix, bool):
        raise ValueError('remove_postfix must be a boolean')
    if speller is not None:
        if not hasattr(speller, 'correct') and not hasattr(
            speller, 'normalize_elongated'
        ):
            raise ValueError(
                'speller must has `correct` or `normalize_elongated` method'
            )

    if expand_hashtags:
        if validate:
            check_file(PATH_PREPROCESSING[1], S3_PATH_PREPROCESSING[1])
        else:
            if not check_available(PATH_PREPROCESSING[1]):
                raise Exception(
                    'preprocessing is not available, please `validate = True`'
                )
        if validate:
            check_file(PATH_PREPROCESSING[2], S3_PATH_PREPROCESSING[2])
        else:
            if not check_available(PATH_PREPROCESSING[2]):
                raise Exception(
                    'preprocessing is not available, please `validate = True`'
                )

    if translate_english_to_bm:
        if validate:
            check_file(
                PATH_PREPROCESSING['english-malay'],
                S3_PATH_PREPROCESSING['english-malay'],
            )
        else:
            if not check_available(PATH_PREPROCESSING['english-malay']):
                raise Exception(
                    'translator english-malay is not available, please `validate = True`'
                )

        with open(PATH_PREPROCESSING['english-malay']['model']) as fopen:
            translator = json.load(fopen)
    else:
        translator = None

    return _Preprocessing(
        normalize = normalize,
        annotate = annotate,
        lowercase = lowercase,
        fix_unidecode = fix_unidecode,
        expand_hashtags = expand_hashtags,
        expand_english_contractions = expand_english_contractions,
        remove_postfix = remove_postfix,
        maxlen_segmenter = maxlen_segmenter,
        translator = translator,
        speller = speller,
    )


def segmenter(max_split_length = 20, validate = True):
    """
    Load Segmenter class.

    Parameters
    ----------
    max_split_length: int, (default=20)
        max length of words in a sentence to segment
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    _Segmenter : malaya.preprocessing._Segmenter class
    """

    if validate:
        check_file(PATH_PREPROCESSING[1], S3_PATH_PREPROCESSING[1])
    else:
        if not check_available(PATH_PREPROCESSING[1]):
            raise Exception(
                'preprocessing is not available, please `validate = True`'
            )
    if validate:
        check_file(PATH_PREPROCESSING[2], S3_PATH_PREPROCESSING[2])
    else:
        if not check_available(PATH_PREPROCESSING[2]):
            raise Exception(
                'preprocessing is not available, please `validate = True`'
            )
    return _Segmenter(max_split_length = max_split_length)


_tokenizer = SocialTokenizer().tokenize



from __future__ import print_function, unicode_literals

BASE = {
    0: [],
    1: ['satu'],
    2: ['dua'],
    3: ['tiga'],
    4: ['empat'],
    5: ['lima'],
    6: ['enam'],
    7: ['tujuh'],
    8: ['lapan'],
    9: ['sembilan'],
}

TENS_TO = {
    3: 'ribu',
    6: 'juta',
    9: 'bilion',
    12: 'trilion',
    15: 'quadrillion',
    18: 'quintillion',
    21: 'sextillion',
    24: 'septillion',
    27: 'oktillion',
    30: 'nonillion',
    33: 'decillion',
}

errmsg_floatord = 'Cannot treat float number as ordinal'
errmsg_negord = 'Cannot treat negative number as ordinal'
errmsg_toobig = 'Too large'
max_num = 10 ** 36


def verify_ordinal(value):
    if not value == int(value):
        raise TypeError(errmsg_floatord % value)
    if not abs(value) == value:
        raise TypeError(errmsg_negord % value)


def split_by_koma(number):
    return str(number).split('.')


def ratus(number):
    if number == '1':
        return ['seratus']
    elif number == '0':
        return []
    else:
        return BASE[int(number)] + ['ratus']


def puluh(number):
    if number[0] == '1':
        if number[1] == '0':
            return ['sepuluh']
        elif number[1] == '1':
            return ['sebelas']
        else:
            return BASE[int(number[1])] + ['belas']
    elif number[0] == '0':
        return BASE[int(number[1])]
    else:
        return BASE[int(number[0])] + ['puluh'] + BASE[int(number[1])]


def split_by_3(number):
    blocks = ()
    length = len(number)
    if length < 3:
        blocks += ((number,),)
    else:
        len_of_first_block = length % 3
        if len_of_first_block > 0:
            blocks += (number[0:len_of_first_block],)
        for i in range(len_of_first_block, length, 3):
            blocks += ((number[i : i + 3],),)
    return blocks


def spell(blocks):
    word_blocks = ()
    first_block = blocks[0]
    if len(first_block[0]) == 1:
        if first_block[0] == '0':
            spelling = ['kosong']
        else:
            spelling = BASE[int(first_block[0])]
    elif len(first_block[0]) == 2:
        spelling = puluh(first_block[0])
    else:
        spelling = ratus(first_block[0][0]) + puluh(first_block[0][1:3])
    word_blocks += ((first_block[0], spelling),)
    for block in blocks[1:]:
        spelling = ratus(block[0][0]) + puluh(block[0][1:3])
        block += (spelling,)
        word_blocks += (block,)
    return word_blocks


def spell_float(float_part):
    word_list = []
    for n in float_part:
        if n == '0':
            word_list += ['kosong']
            continue
        word_list += BASE[int(n)]
    return ' '.join(['', 'perpuluhan'] + word_list)


def join(word_blocks, float_part):
    word_list = []
    length = len(word_blocks) - 1
    first_block = (word_blocks[0],)
    start = 0

    if length == 1 and first_block[0][0] == '1':
        word_list += ['seribu']
        start = 1

    for i in range(start, length + 1, 1):
        word_list += word_blocks[i][1]
        if not word_blocks[i][1]:
            continue
        if i == length:
            break
        word_list += [TENS_TO[(length - i) * 3]]

    return ' '.join(word_list) + float_part


def to_cardinal(number):
    """
    Translate from number input to cardinal text representation

    Parameters
    ----------
    number: int

    Returns
    -------
    string: cardinal representation
    """
    if number >= max_num:
        raise OverflowError(errmsg_toobig % (number, max_num))
    minus = ''
    if number < 0:
        minus = 'negatif '
    float_word = ''
    n = split_by_koma(abs(number))
    if len(n) == 2:
        float_word = spell_float(n[1])
    return minus + join(spell(split_by_3(n[0])), float_word)


def to_ordinal(number):
    """
    Translate from number input to ordinal text representation

    Parameters
    ----------
    number: int

    Returns
    -------
    string: ordinal representation
    """

    verify_ordinal(number)
    out_word = to_cardinal(number)
    if out_word == 'satu':
        return 'pertama'
    return 'ke' + out_word


def to_ordinal_num(number):
    """
    Translate from number input to ordinal numering text representation

    Parameters
    ----------
    number: int

    Returns
    -------
    string: ordinal numering representation
    """

    verify_ordinal(number)
    return 'ke-' + str(number)


def to_currency(value):
    """
    Translate from number input to cardinal currency text representation

    Parameters
    ----------
    number: int

    Returns
    -------
    string: cardinal currency representation
    """

    return to_cardinal(value) + ' ringgit'


def to_year(value):
    """
    Translate from number input to cardinal year text representation

    Parameters
    ----------
    number: int

    Returns
    -------
    string: cardinal year representation
    """
    return to_cardinal(value)



import pickle
import os
import numpy as np
from .texts._jarowrinkler import JaroWinkler
from scipy.spatial.distance import cdist
from sklearn.neighbors import NearestNeighbors
import tensorflow as tf
from . import home
from ._utils._utils import download_file, _Calculator


def load_wiki():
    """
    Return malaya pretrained wikipedia word2vec size 256.

    Returns
    -------
    dictionary: dictionary of dictionary, reverse dictionary and vectors
    """
    if not os.path.isfile('%s/word2vec-wiki/word2vec.p' % (home)):
        print('downloading word2vec-wiki embedded')
        download_file(
            'v13/word2vec/word2vec-wiki-nce-256.p',
            '%s/word2vec-wiki/word2vec.p' % (home),
        )
    with open('%s/word2vec-wiki/word2vec.p' % (home), 'rb') as fopen:
        return pickle.load(fopen)


def load_news(size = 256):
    """
    Return malaya pretrained news word2vec.

    Parameters
    ----------
    size: int, (default=256)

    Returns
    -------
    dictionary: dictionary of dictionary, reverse dictionary and vectors
    """
    if not isinstance(size, int):
        raise ValueError('input must be an integer')
    if size not in [32, 64, 128, 256, 512]:
        raise ValueError(
            'size word2vec not supported, only supports [32, 64, 128, 256, 512]'
        )
    if not os.path.isfile('%s/word2vec-%d/word2vec.p' % (home, size)):
        print('downloading word2vec-%d embedded' % (size))
        download_file(
            'v7/word2vec/word2vec-%d.p' % (size),
            '%s/word2vec-%d/word2vec.p' % (home, size),
        )
    with open('%s/word2vec-%d/word2vec.p' % (home, size), 'rb') as fopen:
        return pickle.load(fopen)


def load(embed_matrix, dictionary):

    """
    Return malaya.wordvector._wordvector object.

    Parameters
    ----------
    embed_matrix: numpy array
    dictionary: dictionary

    Returns
    -------
    _wordvector: malaya.wordvector._wordvector object
    """

    return _wordvector(embed_matrix = embed_matrix, dictionary = dictionary)


class _wordvector:
    def __init__(self, embed_matrix, dictionary):
        self._embed_matrix = embed_matrix
        self._dictionary = dictionary
        self._reverse_dictionary = {v: k for k, v in dictionary.items()}
        self.words = list(dictionary.keys())
        self._jarowinkler = JaroWinkler()
        _graph = tf.Graph()
        with _graph.as_default():
            self._embedding = tf.placeholder(
                tf.float32, self._embed_matrix.shape
            )
            self._x = tf.placeholder(
                tf.float32, [None, self._embed_matrix.shape[1]]
            )
            normed_embedding = tf.nn.l2_normalize(self._embedding, axis = 1)
            normed_array = tf.nn.l2_normalize(self._x, axis = 1)
            self._cosine_similarity = tf.matmul(
                normed_array, tf.transpose(normed_embedding, [1, 0])
            )
            self._sess = tf.InteractiveSession()

    def get_vector_by_name(self, word):
        """
        get vector based on string.

        Parameters
        ----------
        word: str

        Returns
        -------
        vector: numpy
        """
        if not isinstance(word, str):
            raise ValueError('input must be a string')
        if word not in self._dictionary:
            arr = np.array(
                [self._jarowinkler.similarity(word, k) for k in self.words]
            )
            idx = (-arr).argsort()[:5]
            strings = ', '.join([self.words[i] for i in idx])
            raise Exception(
                'input not found in dictionary, here top-5 nearest words [%s]'
                % (strings)
            )
        return self._embed_matrix[self._dictionary[word]]

    def tree_plot(self, labels, figsize = (7, 7), annotate = True):
        """
        plot a tree plot based on output from calculator / n_closest / analogy.

        Parameters
        ----------
        labels : list
            output from calculator / n_closest / analogy.
        visualize : bool
            if True, it will render plt.show, else return data.
        figsize : tuple, (default=(7, 7))
            figure size for plot.

        Returns
        -------
        results: [embed, labelled]
        """

        try:
            import matplotlib.pyplot as plt
            import seaborn as sns

            sns.set()
        except:
            raise Exception(
                'matplotlib and seaborn not installed. Please install it and try again.'
            )

        if not isinstance(labels, list):
            raise ValueError('input must be a list')
        if not isinstance(figsize, tuple):
            raise ValueError('figsize must be a tuple')
        if not isinstance(annotate, bool):
            raise ValueError('annotate must be a boolean')

        idx = [
            self.words.index(e[0] if isinstance(e, list) else e) for e in labels
        ]
        embed = self._embed_matrix[idx]
        embed = embed.dot(embed.T)
        embed = (embed - embed.min()) / (embed.max() - embed.min())
        labelled = []
        for label in labels:
            label = (
                '%s, %.3f' % (label[0], label[1])
                if isinstance(label, list)
                else label
            )
            labelled.append(label)

        plt.figure(figsize = figsize)
        g = sns.clustermap(
            embed,
            cmap = 'Blues',
            xticklabels = labelled,
            yticklabels = labelled,
            annot = annotate,
        )
        plt.show()
        return embed, labelled

    def scatter_plot(
        self,
        labels,
        centre = None,
        figsize = (7, 7),
        plus_minus = 25,
        handoff = 5e-5,
    ):
        """
        plot a scatter plot based on output from calculator / n_closest / analogy.

        Parameters
        ----------
        labels : list
            output from calculator / n_closest / analogy
        centre : str, (default=None)
            centre label, if a str, it will annotate in a red color.
        figsize : tuple, (default=(7, 7))
            figure size for plot.

        Returns
        -------
        list_dictionaries: list of results
        """

        try:
            from sklearn.manifold import TSNE
            import matplotlib.pyplot as plt
            import seaborn as sns

            sns.set()
        except:
            raise Exception(
                'matplotlib and seaborn not installed. Please install it and try again.'
            )

        if not isinstance(labels, list):
            raise ValueError('input must be a list')
        if not isinstance(figsize, tuple):
            raise ValueError('figsize must be a tuple')
        if not isinstance(plus_minus, int):
            raise ValueError('plus_minus must be an integer')

        idx = [
            self.words.index(e[0] if isinstance(e, list) else e) for e in labels
        ]
        if centre:
            if not isinstance(centre, str):
                raise ValueError('centre must be a string')
            idx.append(self.words.index(centre))
        cp_idx = idx[:]
        for i in idx:
            cp_idx.extend(np.arange(i - plus_minus, i).tolist())
            cp_idx.extend(np.arange(i, i + plus_minus).tolist())
        tsne = TSNE(n_components = 2, random_state = 0).fit_transform(
            self._embed_matrix[cp_idx]
        )

        plt.figure(figsize = figsize)
        plt.scatter(tsne[:, 0], tsne[:, 1])
        for label, x, y in zip(
            labels, tsne[: len(labels), 0], tsne[: len(labels), 1]
        ):
            label = (
                '%s, %.3f' % (label[0], label[1])
                if isinstance(label, list)
                else label
            )
            plt.annotate(
                label,
                xy = (x, y),
                xytext = (0, 0),
                textcoords = 'offset points',
            )
        if centre:
            plt.annotate(
                centre,
                xy = (tsne[len(labels), 0], tsne[len(labels), 1]),
                xytext = (0, 0),
                textcoords = 'offset points',
                color = 'red',
            )
        plt.xlim(
            tsne[: len(idx), 0].min() + handoff,
            tsne[: len(idx), 0].max() + handoff,
        )
        plt.ylim(
            tsne[: len(idx), 1].min() + handoff,
            tsne[: len(idx), 1].max() + handoff,
        )
        plt.xticks([])
        plt.yticks([])
        plt.show()
        return tsne

    def _calculate(self, equation):
        tokens, temp = [], ''
        for char in equation:
            if char == ' ':
                continue
            if char not in '()*+-':
                temp += char
            else:
                if len(temp):
                    row = self._dictionary[
                        self.words[
                            np.argmax(
                                [
                                    self._jarowinkler.similarity(temp, k)
                                    for k in self.words
                                ]
                            )
                        ]
                    ]
                    tokens.append(
                        ','.join(
                            self._embed_matrix[row, :].astype('str').tolist()
                        )
                    )
                    temp = ''
                tokens.append(char)
        if len(temp):
            row = self._dictionary[
                self.words[
                    np.argmax(
                        [
                            self._jarowinkler.similarity(temp, k)
                            for k in self.words
                        ]
                    )
                ]
            ]
            tokens.append(
                ','.join(self._embed_matrix[row, :].astype('str').tolist())
            )
        return _Calculator(tokens).exp()

    def _batch_process(self, batch, num_closest = 5, return_similarity = True):
        top_k = tf.nn.top_k(self._cosine_similarity, k = num_closest)
        results = self._sess.run(
            top_k,
            feed_dict = {self._x: batch, self._embedding: self._embed_matrix},
        )
        indices = results.indices
        values = results.values
        words = []
        if not return_similarity:
            for result in indices:
                words.append([self._reverse_dictionary[i] for i in result])
        else:
            for no in range(len(results)):
                words.append(
                    [
                        (
                            self._reverse_dictionary[indices[no, i]],
                            values[no, i],
                        )
                        for i in range(len(indices[no]))
                    ]
                )
        return words

    def batch_calculator(
        self, equations, num_closest = 5, return_similarity = False
    ):
        """
        batch calculator parser for word2vec using tensorflow.

        Parameters
        ----------
        equations: list of str
            Eg, '[(mahathir + najib) - rosmah]'
        num_closest: int, (default=5)
            number of words closest to the result.

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(equations, list):
            raise ValueError('equations must be a list of string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(return_similarity, bool):
            raise ValueError('return_similarity must be a boolean')
        batches = np.array([self._calculate(eq) for eq in equations])
        return self._batch_process(
            batches,
            num_closest = num_closest,
            return_similarity = return_similarity,
        )

    def calculator(
        self,
        equation,
        num_closest = 5,
        metric = 'cosine',
        return_similarity = True,
    ):
        """
        calculator parser for word2vec.

        Parameters
        ----------
        equation: str
            Eg, '(mahathir + najib) - rosmah'
        num_closest: int, (default=5)
            number of words closest to the result.
        metric: str, (default='cosine')
            vector distance algorithm.
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance.

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(equation, str):
            raise ValueError('input must be a string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(metric, str):
            raise ValueError('metric must be a string')
        if not isinstance(return_similarity, bool):
            raise ValueError('num_closest must be a boolean')
        calculated = self._calculate(equation)
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(calculated.reshape((1, -1)))
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            closest_indices = self.closest_row_indices(
                calculated, num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            return word_list

    def batch_n_closest(
        self, words, num_closest = 5, return_similarity = False, soft = True
    ):
        """
        find nearest words based on a batch of words using Tensorflow.

        Parameters
        ----------
        words: list
            Eg, ['najib','anwar']
        num_closest: int, (default=5)
            number of words closest to the result.
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance.
        soft: bool, (default=True)
            if True, a word not in the dictionary will be replaced with nearest JaroWinkler ratio.
            if False, it will throw an exception if a word not in the dictionary.

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(words, list):
            raise ValueError('input must be list of strings')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(return_similarity, bool):
            raise ValueError('return_similarity must be a boolean')
        if not isinstance(soft, bool):
            raise ValueError('soft must be a boolean')
        if soft:
            for i in range(len(words)):
                if words[i] not in self.words:
                    words[i] = self.words[
                        np.argmax(
                            [
                                self._jarowinkler.similarity(words[i], k)
                                for k in self.words
                            ]
                        )
                    ]
        else:
            for i in range(len(words)):
                if words[i] not in self.words:
                    raise Exception(
                        '%s not in dictionary, please use another word or set `soft` = True'
                        % (words[i])
                    )
        batches = np.array([self.get_vector_by_name(w) for w in words])
        return self._batch_process(
            batches,
            num_closest = num_closest,
            return_similarity = return_similarity,
        )

    def n_closest(
        self, word, num_closest = 5, metric = 'cosine', return_similarity = True
    ):
        """
        find nearest words based on a word.

        Parameters
        ----------
        word: str
            Eg, 'najib'
        num_closest: int, (default=5)
            number of words closest to the result.
        metric: str, (default='cosine')
            vector distance algorithm.
        return_similarity: bool, (default=True)
            if True, will return between 0-1 represents the distance.

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(word, str):
            raise ValueError('input must be a string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(metric, str):
            raise ValueError('metric must be a string')
        if not isinstance(return_similarity, bool):
            raise ValueError('num_closest must be a boolean')
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(
                self.get_vector_by_name(word).reshape((1, -1))
            )
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            wv = self.get_vector_by_name(word)
            closest_indices = self.closest_row_indices(
                wv, num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            if word in word_list:
                word_list.remove(word)
            return word_list

    def closest_row_indices(self, wv, num, metric):
        dist_array = np.ravel(
            cdist(self._embed_matrix, wv.reshape((1, -1)), metric = metric)
        )
        sorted_indices = np.argsort(dist_array)
        return sorted_indices[:num]

    def analogy(self, a, b, c, num = 1, metric = 'cosine'):
        """
        analogy calculation, vb - va + vc.

        Parameters
        ----------
        a: str
        b: str
        c: str
        num: int, (default=1)
        metric: str, (default='cosine')
            vector distance algorithm.

        Returns
        -------
        word_list: list of nearest words
        """
        if not isinstance(a, str):
            raise ValueError('a must be a string')
        if not isinstance(b, str):
            raise ValueError('b must be a string')
        if not isinstance(c, str):
            raise ValueError('c must be a string')
        if a not in self._dictionary:
            raise Exception('a not in dictinary')
        if b not in self._dictionary:
            raise Exception('b not in dictinary')
        if c not in self._dictionary:
            raise Exception('c not in dictinary')
        va = self.get_vector_by_name(a)
        vb = self.get_vector_by_name(b)
        vc = self.get_vector_by_name(c)
        vd = vb - va + vc
        closest_indices = self.closest_row_indices(vd, num, metric)
        d_word_list = []
        for i in closest_indices:
            d_word_list.append(self._reverse_dictionary[i])
        return d_word_list

    def project_2d(self, start, end):
        """
        project word2vec into 2d dimension.

        Parameters
        ----------
        start: int
        end: int

        Returns
        -------
        tsne decomposition: numpy
        """
        if not isinstance(start, int):
            raise ValueError('start must be an integer')
        if not isinstance(end, int):
            raise ValueError('end must be an integer')
        tsne = TSNE(n_components = 2)
        embed_2d = tsne.fit_transform(self._embed_matrix[start:end, :])
        word_list = []
        for i in range(start, end):
            word_list.append(self._reverse_dictionary[i])
        return embed_2d, word_list

    def network(
        self,
        word,
        num_closest = 8,
        depth = 4,
        min_distance = 0.5,
        iteration = 300,
        figsize = (15, 15),
        node_color = '#72bbd0',
        node_factor = 50,
    ):

        """
        plot a social network based on word given

        Parameters
        ----------
        word : str
            centre of social network.
        num_closest: int, (default=8)
            number of words closest to the node.
        depth: int, (default=4)
            depth of social network. More deeper more expensive to calculate, big^O(num_closest ** depth).
        min_distance: float, (default=0.5)
            minimum distance among nodes. Increase the value to increase the distance among nodes.
        iteration: int, (default=300)
            number of loops to train the social network to fit min_distace.
        figsize: tuple, (default=(15, 15))
            figure size for plot.
        node_color: str, (default='#72bbd0')
            color for nodes.
        node_factor: int, (default=10)
            size factor for depth nodes. Increase this value will increase nodes sizes based on depth.

        Returns
        -------
        g: networkx graph object
        """

        try:
            import pandas as pd
            import networkx as nx
            import matplotlib.pyplot as plt
        except:
            raise Exception(
                'matplotlib, networkx and pandas not installed. Please install it and try again.'
            )

        if not isinstance(word, str):
            raise ValueError('input must be a string')
        if not isinstance(num_closest, int):
            raise ValueError('num_closest must be an integer')
        if not isinstance(depth, int):
            raise ValueError('depth must be an integer')
        if not isinstance(min_distance, float):
            raise ValueError('min_distance must be a float')
        if not isinstance(iteration, int):
            raise ValueError('iteration must be an integer')
        if not isinstance(figsize, tuple):
            raise ValueError('figsize must be a tuple')
        n_closest = self.n_closest

        def get_follower(
            centre,
            top,
            max_depth = depth,
            current_depth = 0,
            accepted_list = [],
            data = [],
        ):
            if current_depth == max_depth:
                return data, accepted_list
            if centre in accepted_list:
                return data, accepted_list
            else:
                accepted_list.append(centre)

            closest = n_closest(
                centre, num_closest = num_closest, return_similarity = False
            )

            d = {
                'name': centre,
                'followers_ids': closest,
                'centre': top,
                'power': max_depth - current_depth,
            }

            data.append(d)

            cd = current_depth

            if cd + 1 < max_depth:
                for fid in closest:
                    data, accepted_list = get_follower(
                        fid,
                        fid,
                        max_depth = max_depth,
                        current_depth = cd + 1,
                        accepted_list = accepted_list,
                        data = data,
                    )

            return data, accepted_list

        data, accepted_list = get_follower(word, word)
        df = pd.DataFrame(data)
        id_unique = np.unique(df['centre']).tolist()

        followers = []
        for i in range(df.shape[0]):
            followers += df.followers_ids.iloc[i]
        followers = list(set(followers))

        followers = [i for i in followers if i in id_unique]
        true_followers = []
        for i in range(df.shape[0]):
            follows = df.followers_ids.iloc[i]
            follows = [k for k in follows if k in followers]
            true_followers.append(follows)

        df['true_followers'] = true_followers

        G = nx.Graph()

        for i in range(df.shape[0]):
            size = df.power.iloc[i] * node_factor
            G.add_node(df.centre.iloc[i], name = df.centre.iloc[i], size = size)

        for i in range(df.shape[0]):
            for k in df.true_followers.iloc[i]:
                G.add_edge(df.centre.iloc[i], k)

        sizes = [G.node[node]['size'] for node in G]
        names = [G.node[node]['name'] for node in G]

        labeldict = dict(zip(G.nodes(), names))
        plt.figure(figsize = figsize)
        plt.axis('equal')
        nx.draw(
            G,
            node_color = node_color,
            labels = labeldict,
            with_labels = 1,
            node_size = sizes,
            pos = nx.spring_layout(G, k = min_distance, iterations = iteration),
        )
        plt.show()
        return G



# coding=utf-8
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python2, python3
# coding=utf-8
"""Create masked LM/next sentence masked_lm TF examples for ALBERT."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import random

import numpy as np
import six
from six.moves import range
from six.moves import zip
import tensorflow as tf

import tokenization

flags = tf.flags

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'input_file',
    None,
    'Input raw text file (or comma-separated list of files).',
)

flags.DEFINE_string(
    'output_file',
    None,
    'Output TF example file (or comma-separated list of files).',
)

flags.DEFINE_string(
    'vocab_file',
    None,
    'The vocabulary file that the ALBERT model was trained on.',
)

flags.DEFINE_string(
    'spm_model_file', None, 'The model file for sentence piece tokenization.'
)

flags.DEFINE_bool(
    'do_lower_case',
    True,
    'Whether to lower case the input text. Should be True for uncased '
    'models and False for cased models.',
)

flags.DEFINE_bool(
    'do_whole_word_mask',
    True,
    'Whether to use whole word masking rather than per-WordPiece masking.',
)

flags.DEFINE_bool(
    'do_permutation', False, 'Whether to do the permutation training.'
)

flags.DEFINE_bool(
    'favor_shorter_ngram',
    False,
    'Whether to set higher probabilities for sampling shorter ngrams.',
)

flags.DEFINE_bool(
    'random_next_sentence',
    False,
    "Whether to use the sentence that's right before the current sentence "
    'as the negative sample for next sentence prection, rather than using '
    'sentences from other random documents.',
)

flags.DEFINE_integer('max_seq_length', 512, 'Maximum sequence length.')

flags.DEFINE_integer('ngram', 3, 'Maximum number of ngrams to mask.')

flags.DEFINE_integer(
    'max_predictions_per_seq',
    20,
    'Maximum number of masked LM predictions per sequence.',
)

flags.DEFINE_integer('random_seed', 12345, 'Random seed for data generation.')

flags.DEFINE_integer(
    'dupe_factor',
    40,
    'Number of times to duplicate the input data (with different masks).',
)

flags.DEFINE_float('masked_lm_prob', 0.15, 'Masked LM probability.')

flags.DEFINE_float(
    'short_seq_prob',
    0.1,
    'Probability of creating sequences which are shorter than the '
    'maximum length.',
)


class TrainingInstance(object):
    """A single training instance (sentence pair)."""

    def __init__(
        self,
        tokens,
        segment_ids,
        masked_lm_positions,
        masked_lm_labels,
        is_random_next,
        token_boundary,
    ):
        self.tokens = tokens
        self.segment_ids = segment_ids
        self.is_random_next = is_random_next
        self.token_boundary = token_boundary
        self.masked_lm_positions = masked_lm_positions
        self.masked_lm_labels = masked_lm_labels

    def __str__(self):
        s = ''
        s += 'tokens: %s\n' % (
            ' '.join([tokenization.printable_text(x) for x in self.tokens])
        )
        s += 'segment_ids: %s\n' % (
            ' '.join([str(x) for x in self.segment_ids])
        )
        s += 'token_boundary: %s\n' % (
            ' '.join([str(x) for x in self.token_boundary])
        )
        s += 'is_random_next: %s\n' % self.is_random_next
        s += 'masked_lm_positions: %s\n' % (
            ' '.join([str(x) for x in self.masked_lm_positions])
        )
        s += 'masked_lm_labels: %s\n' % (
            ' '.join(
                [tokenization.printable_text(x) for x in self.masked_lm_labels]
            )
        )
        s += '\n'
        return s

    def __repr__(self):
        return self.__str__()


def write_instance_to_example_files(
    instances, tokenizer, max_seq_length, max_predictions_per_seq, output_files
):
    """Create TF example files from `TrainingInstance`s."""
    writers = []
    for output_file in output_files:
        writers.append(tf.python_io.TFRecordWriter(output_file))

    writer_index = 0

    total_written = 0
    for (inst_index, instance) in enumerate(instances):
        input_ids = tokenizer.convert_tokens_to_ids(instance.tokens)
        input_mask = [1] * len(input_ids)
        segment_ids = list(instance.segment_ids)
        token_boundary = list(instance.token_boundary)
        assert len(input_ids) <= max_seq_length

        while len(input_ids) < max_seq_length:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)
            token_boundary.append(0)

        assert len(input_ids) == max_seq_length
        assert len(input_mask) == max_seq_length
        assert len(segment_ids) == max_seq_length

        masked_lm_positions = list(instance.masked_lm_positions)
        masked_lm_ids = tokenizer.convert_tokens_to_ids(
            instance.masked_lm_labels
        )
        masked_lm_weights = [1.0] * len(masked_lm_ids)

        multiplier = 1 + int(FLAGS.do_permutation)
        while len(masked_lm_positions) < max_predictions_per_seq * multiplier:
            masked_lm_positions.append(0)
            masked_lm_ids.append(0)
            masked_lm_weights.append(0.0)

        sentence_order_label = 1 if instance.is_random_next else 0

        features = collections.OrderedDict()
        features['input_ids'] = create_int_feature(input_ids)
        features['input_mask'] = create_int_feature(input_mask)
        features['segment_ids'] = create_int_feature(segment_ids)
        features['token_boundary'] = create_int_feature(token_boundary)
        features['masked_lm_positions'] = create_int_feature(
            masked_lm_positions
        )
        features['masked_lm_ids'] = create_int_feature(masked_lm_ids)
        features['masked_lm_weights'] = create_float_feature(masked_lm_weights)
        # Note: We keep this feature name `next_sentence_labels` to be compatible
        # with the original data created by lanzhzh@. However, in the ALBERT case
        # it does contain sentence_order_label.
        features['next_sentence_labels'] = create_int_feature(
            [sentence_order_label]
        )

        tf_example = tf.train.Example(
            features = tf.train.Features(feature = features)
        )

        writers[writer_index].write(tf_example.SerializeToString())
        writer_index = (writer_index + 1) % len(writers)

        total_written += 1

        if inst_index < 20:
            tf.logging.info('*** Example ***')
            tf.logging.info(
                'tokens: %s'
                % ' '.join(
                    [tokenization.printable_text(x) for x in instance.tokens]
                )
            )

            for feature_name in features.keys():
                feature = features[feature_name]
                values = []
                if feature.int64_list.value:
                    values = feature.int64_list.value
                elif feature.float_list.value:
                    values = feature.float_list.value
                tf.logging.info(
                    '%s: %s'
                    % (feature_name, ' '.join([str(x) for x in values]))
                )

    for writer in writers:
        writer.close()

    tf.logging.info('Wrote %d total instances', total_written)


def create_int_feature(values):
    feature = tf.train.Feature(
        int64_list = tf.train.Int64List(value = list(values))
    )
    return feature


def create_float_feature(values):
    feature = tf.train.Feature(
        float_list = tf.train.FloatList(value = list(values))
    )
    return feature


def create_training_instances(
    input_files,
    tokenizer,
    max_seq_length,
    dupe_factor,
    short_seq_prob,
    masked_lm_prob,
    max_predictions_per_seq,
    rng,
):
    """Create `TrainingInstance`s from raw text."""
    all_documents = [[]]

    # Input file format:
    # (1) One sentence per line. These should ideally be actual sentences, not
    # entire paragraphs or arbitrary spans of text. (Because we use the
    # sentence boundaries for the "next sentence prediction" task).
    # (2) Blank lines between documents. Document boundaries are needed so
    # that the "next sentence prediction" task doesn't span between documents.
    for input_file in input_files:
        with tf.gfile.GFile(input_file, 'r') as reader:
            while True:
                line = reader.readline()
                if not FLAGS.spm_model_file:
                    line = tokenization.convert_to_unicode(line)
                if not line:
                    break
                if FLAGS.spm_model_file:
                    line = tokenization.preprocess_text(
                        line, lower = FLAGS.do_lower_case
                    )
                else:
                    line = line.strip()

                # Empty lines are used as document delimiters
                if not line:
                    all_documents.append([])
                tokens = tokenizer.tokenize(line)
                if tokens:
                    all_documents[-1].append(tokens)

    # Remove empty documents
    all_documents = [x for x in all_documents if x]
    rng.shuffle(all_documents)

    vocab_words = list(tokenizer.vocab.keys())
    instances = []
    for _ in range(dupe_factor):
        for document_index in range(len(all_documents)):
            instances.extend(
                create_instances_from_document(
                    all_documents,
                    document_index,
                    max_seq_length,
                    short_seq_prob,
                    masked_lm_prob,
                    max_predictions_per_seq,
                    vocab_words,
                    rng,
                )
            )

    rng.shuffle(instances)
    return instances


def create_instances_from_document(
    all_documents,
    document_index,
    max_seq_length,
    short_seq_prob,
    masked_lm_prob,
    max_predictions_per_seq,
    vocab_words,
    rng,
):
    """Creates `TrainingInstance`s for a single document."""
    document = all_documents[document_index]

    # Account for [CLS], [SEP], [SEP]
    max_num_tokens = max_seq_length - 3

    # We *usually* want to fill up the entire sequence since we are padding
    # to `max_seq_length` anyways, so short sequences are generally wasted
    # computation. However, we *sometimes*
    # (i.e., short_seq_prob == 0.1 == 10% of the time) want to use shorter
    # sequences to minimize the mismatch between pre-training and fine-tuning.
    # The `target_seq_length` is just a rough target however, whereas
    # `max_seq_length` is a hard limit.
    target_seq_length = max_num_tokens
    if rng.random() < short_seq_prob:
        target_seq_length = rng.randint(2, max_num_tokens)

    # We DON'T just concatenate all of the tokens from a document into a long
    # sequence and choose an arbitrary split point because this would make the
    # next sentence prediction task too easy. Instead, we split the input into
    # segments "A" and "B" based on the actual "sentences" provided by the user
    # input.
    instances = []
    current_chunk = []
    current_length = 0
    i = 0
    while i < len(document):
        segment = document[i]
        current_chunk.append(segment)
        current_length += len(segment)
        if i == len(document) - 1 or current_length >= target_seq_length:
            if current_chunk:
                # `a_end` is how many segments from `current_chunk` go into the `A`
                # (first) sentence.
                a_end = 1
                if len(current_chunk) >= 2:
                    a_end = rng.randint(1, len(current_chunk) - 1)

                tokens_a = []
                for j in range(a_end):
                    tokens_a.extend(current_chunk[j])

                tokens_b = []
                # Random next
                is_random_next = False
                if len(current_chunk) == 1 or (
                    FLAGS.random_next_sentence and rng.random() < 0.5
                ):
                    is_random_next = True
                    target_b_length = target_seq_length - len(tokens_a)

                    # This should rarely go for more than one iteration for large
                    # corpora. However, just to be careful, we try to make sure that
                    # the random document is not the same as the document
                    # we're processing.
                    for _ in range(10):
                        random_document_index = rng.randint(
                            0, len(all_documents) - 1
                        )
                        if random_document_index != document_index:
                            break

                    random_document = all_documents[random_document_index]
                    random_start = rng.randint(0, len(random_document) - 1)
                    for j in range(random_start, len(random_document)):
                        tokens_b.extend(random_document[j])
                        if len(tokens_b) >= target_b_length:
                            break
                    # We didn't actually use these segments so we "put them back" so
                    # they don't go to waste.
                    num_unused_segments = len(current_chunk) - a_end
                    i -= num_unused_segments
                elif not FLAGS.random_next_sentence and rng.random() < 0.5:
                    is_random_next = True
                    for j in range(a_end, len(current_chunk)):
                        tokens_b.extend(current_chunk[j])
                    # Note(mingdachen): in this case, we just swap tokens_a and tokens_b
                    tokens_a, tokens_b = tokens_b, tokens_a
                # Actual next
                else:
                    is_random_next = False
                    for j in range(a_end, len(current_chunk)):
                        tokens_b.extend(current_chunk[j])
                truncate_seq_pair(tokens_a, tokens_b, max_num_tokens, rng)

                assert len(tokens_a) >= 1
                assert len(tokens_b) >= 1

                tokens = []
                segment_ids = []
                tokens.append('[CLS]')
                segment_ids.append(0)
                for token in tokens_a:
                    tokens.append(token)
                    segment_ids.append(0)

                tokens.append('[SEP]')
                segment_ids.append(0)

                for token in tokens_b:
                    tokens.append(token)
                    segment_ids.append(1)
                tokens.append('[SEP]')
                segment_ids.append(1)

                (
                    tokens,
                    masked_lm_positions,
                    masked_lm_labels,
                    token_boundary,
                ) = create_masked_lm_predictions(
                    tokens,
                    masked_lm_prob,
                    max_predictions_per_seq,
                    vocab_words,
                    rng,
                )
                instance = TrainingInstance(
                    tokens = tokens,
                    segment_ids = segment_ids,
                    is_random_next = is_random_next,
                    token_boundary = token_boundary,
                    masked_lm_positions = masked_lm_positions,
                    masked_lm_labels = masked_lm_labels,
                )
                instances.append(instance)
            current_chunk = []
            current_length = 0
        i += 1

    return instances


MaskedLmInstance = collections.namedtuple(
    'MaskedLmInstance', ['index', 'label']
)


def _is_start_piece_sp(piece):
    """Check if the current word piece is the starting piece (sentence piece)."""
    special_pieces = set(list('!"#$%&"()*+,-./:;?@[\\]^_`{|}~'))
    special_pieces.add(u'€'.encode('utf-8'))
    special_pieces.add(u'£'.encode('utf-8'))
    # Note(mingdachen):
    # For foreign characters, we always treat them as a whole piece.
    english_chars = set(list('abcdefghijklmnopqrstuvwhyz'))
    if (
        six.ensure_str(piece).startswith('▁')
        or six.ensure_str(piece).startswith('<')
        or piece in special_pieces
        or not all(
            [
                str(i).lower() in english_chars.union(special_pieces)
                for i in piece
            ]
        )
    ):
        return True
    else:
        return False


def _is_start_piece_bert(piece):
    """Check if the current word piece is the starting piece (BERT)."""
    # When a word has been split into
    # WordPieces, the first token does not have any marker and any subsequence
    # tokens are prefixed with ##. So whenever we see the ## token, we
    # append it to the previous set of word indexes.
    return not six.ensure_str(piece).startswith('##')


def is_start_piece(piece):
    if FLAGS.spm_model_file:
        return _is_start_piece_sp(piece)
    else:
        return _is_start_piece_bert(piece)


def create_masked_lm_predictions(
    tokens, masked_lm_prob, max_predictions_per_seq, vocab_words, rng
):
    """Creates the predictions for the masked LM objective."""

    cand_indexes = []
    # Note(mingdachen): We create a list for recording if the piece is
    # the starting piece of current token, where 1 means true, so that
    # on-the-fly whole word masking is possible.
    token_boundary = [0] * len(tokens)

    for (i, token) in enumerate(tokens):
        if token == '[CLS]' or token == '[SEP]':
            token_boundary[i] = 1
            continue
        # Whole Word Masking means that if we mask all of the wordpieces
        # corresponding to an original word.
        #
        # Note that Whole Word Masking does *not* change the training code
        # at all -- we still predict each WordPiece independently, softmaxed
        # over the entire vocabulary.
        # tf.logging.info(token)
        if (
            FLAGS.do_whole_word_mask
            and len(cand_indexes) >= 1
            and not is_start_piece(token)
        ):
            cand_indexes[-1].append(i)
        else:
            cand_indexes.append([i])
            if is_start_piece(token):
                token_boundary[i] = 1

    output_tokens = list(tokens)

    masked_lm_positions = []
    masked_lm_labels = []

    if masked_lm_prob == 0:
        return (
            output_tokens,
            masked_lm_positions,
            masked_lm_labels,
            token_boundary,
        )

    num_to_predict = min(
        max_predictions_per_seq,
        max(1, int(round(len(tokens) * masked_lm_prob))),
    )

    # Note(mingdachen):
    # By default, we set the probilities to favor longer ngram sequences.
    ngrams = np.arange(1, FLAGS.ngram + 1, dtype = np.int64)
    pvals = 1.0 / np.arange(1, FLAGS.ngram + 1)
    pvals /= pvals.sum(keepdims = True)

    if FLAGS.favor_shorter_ngram:
        pvals = pvals[::-1]

    ngram_indexes = []
    for idx in range(len(cand_indexes)):
        ngram_index = []
        for n in ngrams:
            ngram_index.append(cand_indexes[idx : idx + n])
        ngram_indexes.append(ngram_index)

    rng.shuffle(ngram_indexes)

    masked_lms = []
    covered_indexes = set()
    for cand_index_set in ngram_indexes:
        if len(masked_lms) >= num_to_predict:
            break
        if not cand_index_set:
            continue
        # Note(mingdachen):
        # Skip current piece if they are covered in lm masking or previous ngrams.
        for index_set in cand_index_set[0]:
            for index in index_set:
                if index in covered_indexes:
                    continue

        n = np.random.choice(
            ngrams[: len(cand_index_set)],
            p = pvals[: len(cand_index_set)]
            / pvals[: len(cand_index_set)].sum(keepdims = True),
        )
        index_set = sum(cand_index_set[n - 1], [])
        n -= 1
        # Note(mingdachen):
        # Repeatedly looking for a candidate that does not exceed the
        # maximum number of predictions by trying shorter ngrams.
        while len(masked_lms) + len(index_set) > num_to_predict:
            if n == 0:
                break
            index_set = sum(cand_index_set[n - 1], [])
            n -= 1
        # If adding a whole-word mask would exceed the maximum number of
        # predictions, then just skip this candidate.
        if len(masked_lms) + len(index_set) > num_to_predict:
            continue
        is_any_index_covered = False
        for index in index_set:
            if index in covered_indexes:
                is_any_index_covered = True
                break
        if is_any_index_covered:
            continue
        for index in index_set:
            covered_indexes.add(index)

            masked_token = None
            # 80% of the time, replace with [MASK]
            if rng.random() < 0.8:
                masked_token = '[MASK]'
            else:
                # 10% of the time, keep original
                if rng.random() < 0.5:
                    masked_token = tokens[index]
                # 10% of the time, replace with random word
                else:
                    masked_token = vocab_words[
                        rng.randint(0, len(vocab_words) - 1)
                    ]

            output_tokens[index] = masked_token

            masked_lms.append(
                MaskedLmInstance(index = index, label = tokens[index])
            )
    assert len(masked_lms) <= num_to_predict

    rng.shuffle(ngram_indexes)

    select_indexes = set()
    if FLAGS.do_permutation:
        for cand_index_set in ngram_indexes:
            if len(select_indexes) >= num_to_predict:
                break
            if not cand_index_set:
                continue
            # Note(mingdachen):
            # Skip current piece if they are covered in lm masking or previous ngrams.
            for index_set in cand_index_set[0]:
                for index in index_set:
                    if index in covered_indexes or index in select_indexes:
                        continue

            n = np.random.choice(
                ngrams[: len(cand_index_set)],
                p = pvals[: len(cand_index_set)]
                / pvals[: len(cand_index_set)].sum(keepdims = True),
            )
            index_set = sum(cand_index_set[n - 1], [])
            n -= 1

            while len(select_indexes) + len(index_set) > num_to_predict:
                if n == 0:
                    break
                index_set = sum(cand_index_set[n - 1], [])
                n -= 1
            # If adding a whole-word mask would exceed the maximum number of
            # predictions, then just skip this candidate.
            if len(select_indexes) + len(index_set) > num_to_predict:
                continue
            is_any_index_covered = False
            for index in index_set:
                if index in covered_indexes or index in select_indexes:
                    is_any_index_covered = True
                    break
            if is_any_index_covered:
                continue
            for index in index_set:
                select_indexes.add(index)
        assert len(select_indexes) <= num_to_predict

        select_indexes = sorted(select_indexes)
        permute_indexes = list(select_indexes)
        rng.shuffle(permute_indexes)
        orig_token = list(output_tokens)

        for src_i, tgt_i in zip(select_indexes, permute_indexes):
            output_tokens[src_i] = orig_token[tgt_i]
            masked_lms.append(
                MaskedLmInstance(index = src_i, label = orig_token[src_i])
            )

    masked_lms = sorted(masked_lms, key = lambda x: x.index)

    for p in masked_lms:
        masked_lm_positions.append(p.index)
        masked_lm_labels.append(p.label)
    return (
        output_tokens,
        masked_lm_positions,
        masked_lm_labels,
        token_boundary,
    )


def truncate_seq_pair(tokens_a, tokens_b, max_num_tokens, rng):
    """Truncates a pair of sequences to a maximum sequence length."""
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_num_tokens:
            break

        trunc_tokens = tokens_a if len(tokens_a) > len(tokens_b) else tokens_b
        assert len(trunc_tokens) >= 1

        # We want to sometimes truncate from the front and sometimes from the
        # back to add more randomness and avoid biases.
        if rng.random() < 0.5:
            del trunc_tokens[0]
        else:
            trunc_tokens.pop()


def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)

    tokenizer = tokenization.FullTokenizer(
        vocab_file = FLAGS.vocab_file,
        do_lower_case = FLAGS.do_lower_case,
        spm_model_file = FLAGS.spm_model_file,
    )

    input_files = []
    for input_pattern in FLAGS.input_file.split(','):
        input_files.extend(tf.gfile.Glob(input_pattern))

    tf.logging.info('*** Reading from input files ***')
    for input_file in input_files:
        tf.logging.info('  %s', input_file)

    rng = random.Random(FLAGS.random_seed)
    instances = create_training_instances(
        input_files,
        tokenizer,
        FLAGS.max_seq_length,
        FLAGS.dupe_factor,
        FLAGS.short_seq_prob,
        FLAGS.masked_lm_prob,
        FLAGS.max_predictions_per_seq,
        rng,
    )

    tf.logging.info('number of instances: %i', len(instances))

    output_files = FLAGS.output_file.split(',')
    tf.logging.info('*** Writing to output files ***')
    for output_file in output_files:
        tf.logging.info('  %s', output_file)

    write_instance_to_example_files(
        instances,
        tokenizer,
        FLAGS.max_seq_length,
        FLAGS.max_predictions_per_seq,
        output_files,
    )


if __name__ == '__main__':
    flags.mark_flag_as_required('input_file')
    flags.mark_flag_as_required('output_file')
    flags.mark_flag_as_required('vocab_file')
    tf.app.run()



import tensorflow as tf
from bert import modeling
from ..texts._text_functions import (
    bert_tokenization,
    padding_sequence,
    merge_sentencepiece_tokens,
    merge_wordpiece_tokens,
)
from ._sampling import top_k_logits, top_p_logits
from collections import defaultdict
import numpy as np
import os

bert_num_layers = {'multilanguage': 12, 'base': 12, 'small': 6}


def _extract_attention_weights(num_layers, tf_graph):
    attns = [
        {
            'layer_%s'
            % i: tf_graph.get_tensor_by_name(
                'bert/encoder/layer_%s/attention/self/Softmax:0' % i
            )
        }
        for i in range(num_layers)
    ]

    return attns


def _extract_attention_weights_import(num_layers, tf_graph):
    attns = [
        {
            'layer_%s'
            % i: tf_graph.get_tensor_by_name(
                'import/bert/encoder/layer_%s/attention/self/Softmax:0' % i
            )
        }
        for i in range(num_layers)
    ]

    return attns


class _Model:
    def __init__(self, bert_config, tokenizer, cls, sep):
        _graph = tf.Graph()
        with _graph.as_default():
            self.X = tf.placeholder(tf.int32, [None, None])
            self.top_p = tf.placeholder(tf.float32, None)
            self.top_k = tf.placeholder(tf.int32, None)
            self.k = tf.placeholder(tf.int32, None)
            self.temperature = tf.placeholder(tf.float32, None)
            self.indices = tf.placeholder(tf.int32, [None, None])
            self._tokenizer = tokenizer
            self._cls = cls
            self._sep = sep

            self.model = modeling.BertModel(
                config = bert_config,
                is_training = False,
                input_ids = self.X,
                use_one_hot_embeddings = False,
            )
            self.logits = self.model.get_pooled_output()
            output_layer = self.model.get_sequence_output()
            embedding = self.model.get_embedding_table()

            with tf.variable_scope('cls/predictions'):
                with tf.variable_scope('transform'):
                    input_tensor = tf.layers.dense(
                        output_layer,
                        units = bert_config.hidden_size,
                        activation = modeling.get_activation(
                            bert_config.hidden_act
                        ),
                        kernel_initializer = modeling.create_initializer(
                            bert_config.initializer_range
                        ),
                    )
                    input_tensor = modeling.layer_norm(input_tensor)
                output_bias = tf.get_variable(
                    'output_bias',
                    shape = [bert_config.vocab_size],
                    initializer = tf.zeros_initializer(),
                )
                logits = tf.matmul(input_tensor, embedding, transpose_b = True)
                self._logits = tf.nn.bias_add(logits, output_bias)
                self._log_softmax = tf.nn.log_softmax(self._logits)

            logits = tf.gather_nd(self._logits, self.indices)
            logits = logits / self.temperature

            def necleus():
                return top_p_logits(logits, self.top_p)

            def select_k():
                return top_k_logits(logits, self.top_k)

            logits = tf.cond(self.top_p > 0, necleus, select_k)
            self.samples = tf.multinomial(
                logits, num_samples = self.k, output_dtype = tf.int32
            )

            self._sess = tf.InteractiveSession()
            self._sess.run(tf.global_variables_initializer())
            var_lists = tf.get_collection(
                tf.GraphKeys.TRAINABLE_VARIABLES, scope = 'bert'
            )
            cls = tf.get_collection(
                tf.GraphKeys.TRAINABLE_VARIABLES, scope = 'cls'
            )
            self._saver = tf.train.Saver(var_list = var_lists + cls)
            attns = _extract_attention_weights(
                bert_config.num_hidden_layers, tf.get_default_graph()
            )
            self.attns = attns

    def _log_vectorize(self, s_tokens):

        """
        Log vectorize ids, suitable for spelling correction or any minimizing log probability.

        Parameters
        ----------
        s_tokens : list of tokenized word after sentencepiece.

        Returns
        -------
        array: vectorized strings
        """

        return self._sess.run(self._log_softmax, feed_dict = {self.X: s_tokens})

    def vectorize(self, strings):

        """
        Vectorize string inputs using bert attention.

        Parameters
        ----------
        strings : str / list of str

        Returns
        -------
        array: vectorized strings
        """

        if isinstance(strings, list):
            if not isinstance(strings[0], str):
                raise ValueError('input must be a list of strings or a string')
        else:
            if not isinstance(strings, str):
                raise ValueError('input must be a list of strings or a string')
        if isinstance(strings, str):
            strings = [strings]

        batch_x, _, _, _ = bert_tokenization(
            self._tokenizer, strings, cls = self._cls, sep = self._sep
        )
        return self._sess.run(self.logits, feed_dict = {self.X: batch_x})

    def _attention(self, strings):
        batch_x, _, _, s_tokens = bert_tokenization(
            self._tokenizer, strings, cls = self._cls, sep = self._sep
        )
        maxlen = max([len(s) for s in s_tokens])
        s_tokens = padding_sequence(s_tokens, maxlen, pad_int = self._sep)
        attentions = self._sess.run(self.attns, feed_dict = {self.X: batch_x})
        return attentions, s_tokens

    def attention(self, strings, method = 'last', **kwargs):
        """
        Get attention string inputs from bert attention.

        Parameters
        ----------
        strings : str / list of str
        method : str, optional (default='last')
            Attention layer supported. Allowed values:

            * ``'last'`` - attention from last layer.
            * ``'first'`` - attention from first layer.
            * ``'mean'`` - average attentions from all layers.

        Returns
        -------
        array: attention
        """

        if isinstance(strings, list):
            if not isinstance(strings[0], str):
                raise ValueError('input must be a list of strings or a string')
        else:
            if not isinstance(strings, str):
                raise ValueError('input must be a list of strings or a string')
        if isinstance(strings, str):
            strings = [strings]

        method = method.lower()
        if method not in ['last', 'first', 'mean']:
            raise Exception(
                "method not supported, only support 'last', 'first' and 'mean'"
            )
        attentions, s_tokens = self._attention(strings)

        if method == 'first':
            cls_attn = list(attentions[0].values())[0][:, :, 0, :]

        if method == 'last':
            cls_attn = list(attentions[-1].values())[0][:, :, 0, :]

        if method == 'mean':
            combined_attentions = []
            for a in attentions:
                combined_attentions.append(list(a.values())[0])
            cls_attn = np.mean(combined_attentions, axis = 0).mean(axis = 2)

        cls_attn = np.mean(cls_attn, axis = 1)
        total_weights = np.sum(cls_attn, axis = -1, keepdims = True)
        attn = cls_attn / total_weights
        output = []
        for i in range(attn.shape[0]):
            if '[' in self._cls:
                output.append(
                    merge_wordpiece_tokens(list(zip(s_tokens[i], attn[i])))
                )
            else:
                output.append(
                    merge_sentencepiece_tokens(list(zip(s_tokens[i], attn[i])))
                )
        return output

    def visualize_attention(self, string):
        from .._utils._html import _attention

        if not isinstance(string, str):
            raise ValueError('input must be a string')
        strings = [string]
        attentions, s_tokens = self._attention(strings)
        attn_dict = defaultdict(list)
        for layer, attn_data in enumerate(attentions):
            attn = list(attn_data.values())[0][0]
            attn_dict['all'].append(attn.tolist())

        results = {
            'all': {
                'attn': attn_dict['all'],
                'left_text': s_tokens[0],
                'right_text': s_tokens[0],
            }
        }
        _attention(results)


def available_bert_model():
    """
    List available bert models.
    """
    return ['multilanguage', 'base', 'small']


def bert(model = 'base', lite = False, validate = True):
    """
    Load bert model.

    Parameters
    ----------
    model : str, optional (default='base')
        Model architecture supported. Allowed values:

        * ``'base'`` - base bert-bahasa released by Malaya.
        * ``'small'`` - small bert-bahasa released by Malaya.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    BERT_MODEL: malaya._transformer._bert._Model class
    """

    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    from .._utils._paths import PATH_BERT, S3_PATH_BERT
    from .._utils._utils import check_file, check_available

    model = model.lower()
    if validate:
        check_file(PATH_BERT[model]['model'], S3_PATH_BERT[model])
    else:
        if not check_available(PATH_BERT[model]['model']):
            raise Exception(
                'bert-model/%s is not available, please `validate = True`'
                % (model)
            )
    if not os.path.exists(PATH_BERT[model]['directory']):
        import tarfile

        with tarfile.open(PATH_BERT[model]['model']['model']) as tar:
            tar.extractall(path = PATH_BERT[model]['path'])

    import sentencepiece as spm
    from ..texts._text_functions import SentencePieceTokenizer

    bert_checkpoint = PATH_BERT[model]['directory'] + 'model.ckpt'
    vocab_model = PATH_BERT[model]['directory'] + 'sp10m.cased.v8.model'
    vocab = PATH_BERT[model]['directory'] + 'sp10m.cased.v8.vocab'
    bert_config = PATH_BERT[model]['directory'] + 'config.json'

    sp_model = spm.SentencePieceProcessor()
    sp_model.Load(vocab_model)

    with open(vocab) as fopen:
        v = fopen.read().split('\n')[:-1]
    v = [i.split('\t') for i in v]
    v = {i[0]: i[1] for i in v}
    tokenizer = SentencePieceTokenizer(v, sp_model)
    cls = '<cls>'
    sep = '<sep>'

    bert_config = modeling.BertConfig.from_json_file(bert_config)
    model = _Model(bert_config, tokenizer, cls = cls, sep = sep)
    model._saver.restore(model._sess, bert_checkpoint)
    return model



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os
import tensorflow as tf
from . import modeling


def _get_initializer(FLAGS):
    """Get variable intializer."""
    if FLAGS.init == 'uniform':
        initializer = tf.initializers.random_uniform(
            minval = -FLAGS.init_range, maxval = FLAGS.init_range, seed = None
        )
    elif FLAGS.init == 'normal':
        initializer = tf.initializers.random_normal(
            stddev = FLAGS.init_std, seed = None
        )
    else:
        raise ValueError('Initializer {} not supported'.format(FLAGS.init))
    return initializer


class XLNetConfig(object):
    """XLNetConfig contains hyperparameters that are specific to a model checkpoint;
  i.e., these hyperparameters should be the same between
  pretraining and finetuning.

  The following hyperparameters are defined:
    n_layer: int, the number of layers.
    d_model: int, the hidden size.
    n_head: int, the number of attention heads.
    d_head: int, the dimension size of each attention head.
    d_inner: int, the hidden size in feed-forward layers.
    ff_activation: str, "relu" or "gelu".
    untie_r: bool, whether to untie the biases in attention.
    n_token: int, the vocab size.
  """

    def __init__(self, FLAGS = None, json_path = None):
        """Constructing an XLNetConfig.
    One of FLAGS or json_path should be provided."""

        assert FLAGS is not None or json_path is not None

        self.keys = [
            'n_layer',
            'd_model',
            'n_head',
            'd_head',
            'd_inner',
            'ff_activation',
            'untie_r',
            'n_token',
        ]

        if FLAGS is not None:
            self.init_from_flags(FLAGS)

        if json_path is not None:
            self.init_from_json(json_path)

    def init_from_flags(self, FLAGS):
        for key in self.keys:
            setattr(self, key, getattr(FLAGS, key))

    def init_from_json(self, json_path):
        with tf.gfile.Open(json_path) as f:
            json_data = json.load(f)
            for key in self.keys:
                setattr(self, key, json_data[key])

    def to_json(self, json_path):
        """Save XLNetConfig to a json file."""
        json_data = {}
        for key in self.keys:
            json_data[key] = getattr(self, key)

        json_dir = os.path.dirname(json_path)
        if not tf.gfile.Exists(json_dir):
            tf.gfile.MakeDirs(json_dir)
        with tf.gfile.Open(json_path, 'w') as f:
            json.dump(json_data, f, indent = 4, sort_keys = True)


def create_run_config(is_training, is_finetune, FLAGS):
    kwargs = dict(
        is_training = is_training,
        use_tpu = FLAGS.use_tpu,
        use_bfloat16 = FLAGS.use_bfloat16,
        dropout = FLAGS.dropout,
        dropatt = FLAGS.dropatt,
        init = FLAGS.init,
        init_range = FLAGS.init_range,
        init_std = FLAGS.init_std,
        clamp_len = FLAGS.clamp_len,
    )

    if not is_finetune:
        kwargs.update(
            dict(
                mem_len = FLAGS.mem_len,
                reuse_len = FLAGS.reuse_len,
                bi_data = FLAGS.bi_data,
                clamp_len = FLAGS.clamp_len,
                same_length = FLAGS.same_length,
            )
        )

    return RunConfig(**kwargs)


class RunConfig(object):
    """RunConfig contains hyperparameters that could be different
  between pretraining and finetuning.
  These hyperparameters can also be changed from run to run.
  We store them separately from XLNetConfig for flexibility.
  """

    def __init__(
        self,
        is_training,
        use_tpu,
        use_bfloat16,
        dropout,
        dropatt,
        init = 'normal',
        init_range = 0.1,
        init_std = 0.02,
        mem_len = None,
        reuse_len = None,
        bi_data = False,
        clamp_len = -1,
        same_length = False,
    ):
        """
    Args:
      is_training: bool, whether in training mode.
      use_tpu: bool, whether TPUs are used.
      use_bfloat16: bool, use bfloat16 instead of float32.
      dropout: float, dropout rate.
      dropatt: float, dropout rate on attention probabilities.
      init: str, the initialization scheme, either "normal" or "uniform".
      init_range: float, initialize the parameters with a uniform distribution
        in [-init_range, init_range]. Only effective when init="uniform".
      init_std: float, initialize the parameters with a normal distribution
        with mean 0 and stddev init_std. Only effective when init="normal".
      mem_len: int, the number of tokens to cache.
      reuse_len: int, the number of tokens in the currect batch to be cached
        and reused in the future.
      bi_data: bool, whether to use bidirectional input pipeline.
        Usually set to True during pretraining and False during finetuning.
      clamp_len: int, clamp all relative distances larger than clamp_len.
        -1 means no clamping.
      same_length: bool, whether to use the same attention length for each token.
    """

        self.init = init
        self.init_range = init_range
        self.init_std = init_std
        self.is_training = is_training
        self.dropout = dropout
        self.dropatt = dropatt
        self.use_tpu = use_tpu
        self.use_bfloat16 = use_bfloat16
        self.mem_len = mem_len
        self.reuse_len = reuse_len
        self.bi_data = bi_data
        self.clamp_len = clamp_len
        self.same_length = same_length


class XLNetModel(object):
    """A wrapper of the XLNet model used during both pretraining and finetuning."""

    def __init__(
        self,
        xlnet_config,
        run_config,
        input_ids,
        seg_ids,
        input_mask,
        mems = None,
        perm_mask = None,
        target_mapping = None,
        inp_q = None,
        **kwargs
    ):
        """
    Args:
      xlnet_config: XLNetConfig,
      run_config: RunConfig,
      input_ids: int32 Tensor in shape [len, bsz], the input token IDs.
      seg_ids: int32 Tensor in shape [len, bsz], the input segment IDs.
      input_mask: float32 Tensor in shape [len, bsz], the input mask.
        0 for real tokens and 1 for padding.
      mems: a list of float32 Tensors in shape [mem_len, bsz, d_model], memory
        from previous batches. The length of the list equals n_layer.
        If None, no memory is used.
      perm_mask: float32 Tensor in shape [len, len, bsz].
        If perm_mask[i, j, k] = 0, i attend to j in batch k;
        if perm_mask[i, j, k] = 1, i does not attend to j in batch k.
        If None, each position attends to all the others.
      target_mapping: float32 Tensor in shape [num_predict, len, bsz].
        If target_mapping[i, j, k] = 1, the i-th predict in batch k is
        on the j-th token.
        Only used during pretraining for partial prediction.
        Set to None during finetuning.
      inp_q: float32 Tensor in shape [len, bsz].
        1 for tokens with losses and 0 for tokens without losses.
        Only used during pretraining for two-stream attention.
        Set to None during finetuning.
    """

        initializer = _get_initializer(run_config)

        tfm_args = dict(
            n_token = xlnet_config.n_token,
            initializer = initializer,
            attn_type = 'bi',
            n_layer = xlnet_config.n_layer,
            d_model = xlnet_config.d_model,
            n_head = xlnet_config.n_head,
            d_head = xlnet_config.d_head,
            d_inner = xlnet_config.d_inner,
            ff_activation = xlnet_config.ff_activation,
            untie_r = xlnet_config.untie_r,
            is_training = run_config.is_training,
            use_bfloat16 = run_config.use_bfloat16,
            use_tpu = run_config.use_tpu,
            dropout = run_config.dropout,
            dropatt = run_config.dropatt,
            mem_len = run_config.mem_len,
            reuse_len = run_config.reuse_len,
            bi_data = run_config.bi_data,
            clamp_len = run_config.clamp_len,
            same_length = run_config.same_length,
        )

        input_args = dict(
            inp_k = input_ids,
            seg_id = seg_ids,
            input_mask = input_mask,
            mems = mems,
            perm_mask = perm_mask,
            target_mapping = target_mapping,
            inp_q = inp_q,
        )
        tfm_args.update(input_args)

        with tf.variable_scope('model', reuse = tf.AUTO_REUSE):
            (
                self.output,
                self.new_mems,
                self.lookup_table,
            ) = modeling.transformer_xl(**tfm_args)

        self.input_mask = input_mask
        self.initializer = initializer
        self.xlnet_config = xlnet_config
        self.run_config = run_config

    def get_pooled_out(self, summary_type, use_summ_proj = True):
        """
    Args:
      summary_type: str, "last", "first", "mean", or "attn". The method
        to pool the input to get a vector representation.
      use_summ_proj: bool, whether to use a linear projection during pooling.

    Returns:
      float32 Tensor in shape [bsz, d_model], the pooled representation.
    """

        xlnet_config = self.xlnet_config
        run_config = self.run_config

        with tf.variable_scope('model', reuse = tf.AUTO_REUSE):
            summary = modeling.summarize_sequence(
                summary_type = summary_type,
                hidden = self.output,
                d_model = xlnet_config.d_model,
                n_head = xlnet_config.n_head,
                d_head = xlnet_config.d_head,
                dropout = run_config.dropout,
                dropatt = run_config.dropatt,
                is_training = run_config.is_training,
                input_mask = self.input_mask,
                initializer = self.initializer,
                use_proj = use_summ_proj,
            )

        return summary

    def get_sequence_output(self):
        """
    Returns:
      float32 Tensor in shape [len, bsz, d_model]. The last layer hidden
      representation of XLNet.
    """

        return self.output

    def get_new_memory(self):
        """
    Returns:
      list of float32 Tensors in shape [mem_len, bsz, d_model], the new
      memory that concatenates the previous memory with the current input
      representations.
      The length of the list equals n_layer.
    """
        return self.new_mems

    def get_embedding_table(self):
        """
    Returns:
      float32 Tensor in shape [n_token, d_model]. The embedding lookup table.
      Used for tying embeddings between input and output layers.
    """
        return self.lookup_table

    def get_initializer(self):
        """
    Returns:
      A tf initializer. Used to initialize variables in layers on top of XLNet.
    """
        return self.initializer



import dateparser
import re
from ._text_functions import multireplace
from ._tatabahasa import date_replace
from ._regex import (
    _money,
    _temperature,
    _distance,
    _volume,
    _duration,
    _date,
    _past_date_string,
    _now_date_string,
    _future_date_string,
    _yesterday_tomorrow_date_string,
    _depan_date_string,
    _expressions,
    _left_datetime,
    _right_datetime,
    _today_time,
    _left_datetodaytime,
    _right_datetodaytime,
    _left_yesterdaydatetime,
    _right_yesterdaydatetime,
    _left_yesterdaydatetodaytime,
    _right_yesterdaydatetodaytime,
)
from ._food import (
    hot_ice_beverage_regex,
    fruit_juice_regex,
    unique_beverage_regex,
    total_foods_regex,
)
from ._normalization import money, _normalize_money
from ..cluster import cluster_tagging


class _Entity_regex:
    def __init__(self, model = None):
        self._model = model

    def predict(self, string):
        """
        classify a string.

        Parameters
        ----------
        string : str
        """

        if not isinstance(string, str):
            raise ValueError('input must be a string')

        if self._model:
            result_model = self._model.predict(string)
            result = cluster_tagging(result_model)
        else:
            result = {}
        money_ = re.findall(_money, string)
        money_ = [(s, money(s)[1]) for s in money_]
        dates_ = re.findall(_date, string)
        past_date_string_ = re.findall(_past_date_string, string)
        now_date_string_ = re.findall(_now_date_string, string)
        future_date_string_ = re.findall(_future_date_string, string)
        yesterday_date_string_ = re.findall(
            _yesterday_tomorrow_date_string, string
        )
        depan_date_string_ = re.findall(_depan_date_string, string)
        dates_ = (
            dates_
            + past_date_string_
            + now_date_string_
            + future_date_string_
            + yesterday_date_string_
            + depan_date_string_
        )
        dates_ = [multireplace(s, date_replace) for s in dates_]
        dates_ = [re.sub(r'[ ]+', ' ', s).strip() for s in dates_]
        dates_ = {s: dateparser.parse(s) for s in dates_}
        money_ = {s[0]: s[1] for s in money_}
        temperature_ = re.findall(_temperature, string)
        temperature_ = [re.sub(r'[ ]+', ' ', s).strip() for s in temperature_]
        distance_ = re.findall(_distance, string)
        distance_ = [re.sub(r'[ ]+', ' ', s).strip() for s in distance_]
        volume_ = re.findall(_volume, string)
        volume_ = [re.sub(r'[ ]+', ' ', s).strip() for s in volume_]
        duration_ = re.findall(_duration, string)
        duration_ = [re.sub(r'[ ]+', ' ', s).strip() for s in duration_]
        phone_ = re.findall(_expressions['phone'], string)
        phone_ = [re.sub(r'[ ]+', ' ', s).strip() for s in phone_]
        email_ = re.findall(_expressions['email'], string)
        email_ = [re.sub(r'[ ]+', ' ', s).strip() for s in email_]
        url_ = re.findall(_expressions['url'], string)
        url_ = [re.sub(r'[ ]+', ' ', s).strip() for s in url_]
        time_ = re.findall(_expressions['time'], string)
        time_ = [re.sub(r'[ ]+', ' ', s).strip() for s in time_]
        today_time_ = re.findall(_today_time, string)
        today_time_ = [re.sub(r'[ ]+', ' ', s).strip() for s in today_time_]
        time_ = time_ + today_time_

        if 'time' in result:
            time_ = list(set(time_ + result['time']))

        time_ = [multireplace(s, date_replace) for s in time_]
        time_ = [re.sub(r'[ ]+', ' ', s).strip() for s in time_]
        time_ = {s: dateparser.parse(s) for s in time_}

        left_datetime_ = [
            '%s %s' % (i[0], i[1]) for i in re.findall(_left_datetime, string)
        ]
        right_datetime_ = [
            '%s %s' % (i[0], i[1]) for i in re.findall(_right_datetime, string)
        ]
        today_left_datetime_ = [
            '%s %s' % (i[0], i[1])
            for i in re.findall(_left_datetodaytime, string)
        ]
        today_right_datetime_ = [
            '%s %s' % (i[0], i[1])
            for i in re.findall(_right_datetodaytime, string)
        ]
        left_yesterdaydatetime_ = [
            '%s %s' % (i[0], i[1])
            for i in re.findall(_left_yesterdaydatetime, string)
        ]
        right_yesterdaydatetime_ = [
            '%s %s' % (i[0], i[1])
            for i in re.findall(_right_yesterdaydatetime, string)
        ]
        left_yesterdaydatetodaytime_ = [
            '%s %s' % (i[0], i[1])
            for i in re.findall(_left_yesterdaydatetodaytime, string)
        ]
        right_yesterdaydatetodaytime_ = [
            '%s %s' % (i[0], i[1])
            for i in re.findall(_right_yesterdaydatetodaytime, string)
        ]

        datetime_ = (
            left_datetime_
            + right_datetime_
            + today_left_datetime_
            + today_right_datetime_
            + left_yesterdaydatetime_
            + right_yesterdaydatetime_
            + left_yesterdaydatetodaytime_
            + right_yesterdaydatetodaytime_
        )
        datetime_ = [re.sub(r'[ ]+', ' ', s).strip() for s in datetime_]
        datetime_ = {s: dateparser.parse(s) for s in datetime_}

        foods_ = re.findall(total_foods_regex, string)
        foods_ = [re.sub(r'[ ]+', ' ', s).strip() for s in foods_]

        drinks_ = re.findall(unique_beverage_regex, string)
        drinks_.extend(re.findall(hot_ice_beverage_regex, string))
        drinks_.extend(re.findall(fruit_juice_regex, string))
        drinks_ = [re.sub(r'[ ]+', ' ', s).strip() for s in drinks_]

        result['date'] = dates_
        result['money'] = money_
        result['temperature'] = temperature_
        result['distance'] = distance_
        result['volume'] = volume_
        result['duration'] = duration_
        result['phone'] = phone_
        result['email'] = email_
        result['url'] = url_
        result['time'] = time_
        result['datetime'] = datetime_
        result['food'] = foods_
        result['drink'] = drinks_
        return result



import tensorflow as tf


def top_k_logits(logits, k):
    if k == 0:
        return logits

    def _top_k():
        values, _ = tf.nn.top_k(logits, k = k)
        min_values = values[:, -1, tf.newaxis]
        return tf.where(
            logits < min_values,
            tf.ones_like(logits, dtype = logits.dtype) * -1e10,
            logits,
        )

    return tf.cond(tf.equal(k, 0), lambda: logits, lambda: _top_k())


def top_p_logits(logits, p):
    with tf.variable_scope('top_p_logits'):
        logits_sort = tf.sort(logits, direction = 'DESCENDING')
        probs_sort = tf.nn.softmax(logits_sort)
        probs_sums = tf.cumsum(probs_sort, axis = 1, exclusive = True)
        logits_masked = tf.where(
            probs_sums < p, logits_sort, tf.ones_like(logits_sort) * 1000
        )  # [batchsize, vocab]
        min_logits = tf.reduce_min(
            logits_masked, axis = 1, keepdims = True
        )  # [batchsize, 1]
        return tf.where(
            logits < min_logits,
            tf.ones_like(logits, dtype = logits.dtype) * -1e10,
            logits,
        )



"""Pretraining on GPUs."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os, sys

os.environ['CUDA_VISIBLE_DEVICES'] = '3'
import math
import json
import time
import numpy as np

from absl import flags
import absl.logging as _logging

import tensorflow as tf

import data_utils
import model_utils
from gpu_utils import assign_to_gpu, average_grads_and_vars
from tensorflow.python.distribute.cross_device_ops import (
    AllReduceCrossDeviceOps,
)
import custom_function_builder
from tensorflow.python.estimator.run_config import RunConfig
from tensorflow.python.estimator.estimator import Estimator

flags.DEFINE_integer('num_hosts', default = 1, help = 'Number of hosts')
flags.DEFINE_integer(
    'num_core_per_host', default = 1, help = 'Number of cores per host'
)
flags.DEFINE_bool(
    'use_tpu', default = False, help = 'Whether to use TPUs for training.'
)

flags.DEFINE_integer(
    'num_gpu_cores',
    2,
    'Only used if `use_gpu` is True. Total number of GPU cores to use.',
)

# Experiment (data/checkpoint/directory) config
flags.DEFINE_integer(
    'num_passes', default = 1, help = 'Number of passed used for training.'
)
flags.DEFINE_string(
    'record_info_dir',
    default = None,
    help = 'Path to local directory containing `record_info-lm.json`.',
)
flags.DEFINE_string('model_dir', default = None, help = 'Estimator model_dir.')
flags.DEFINE_string(
    'init_checkpoint',
    default = None,
    help = 'checkpoint path for initializing the model.',
)

# Optimization config
flags.DEFINE_float(
    'learning_rate', default = 1e-4, help = 'Maximum learning rate.'
)
flags.DEFINE_float('clip', default = 1.0, help = 'Gradient clipping value.')
# for cosine decay
flags.DEFINE_float(
    'min_lr_ratio', default = 0.001, help = 'Minimum ratio learning rate.'
)
flags.DEFINE_integer(
    'warmup_steps', default = 0, help = 'Number of steps for linear lr warmup.'
)
flags.DEFINE_float('adam_epsilon', default = 1e-8, help = 'Adam epsilon')
flags.DEFINE_string('decay_method', default = 'poly', help = 'poly or cos')
flags.DEFINE_float('weight_decay', default = 0.0, help = 'weight decay')

# Training config
flags.DEFINE_integer(
    'train_batch_size', default = 16, help = 'Size of train batch.'
)
flags.DEFINE_integer(
    'train_steps', default = 100000, help = 'Total number of training steps.'
)
flags.DEFINE_integer(
    'iterations', default = 1000, help = 'Number of iterations per repeat loop.'
)
flags.DEFINE_integer(
    'save_steps',
    default = None,
    help = 'number of steps for model checkpointing.',
)

# Data config
flags.DEFINE_integer(
    'seq_len', default = 0, help = 'Sequence length for pretraining.'
)
flags.DEFINE_integer(
    'reuse_len',
    default = 0,
    help = 'How many tokens to be reused in the next batch. '
    'Could be half of seq_len',
)
flags.DEFINE_bool(
    'bi_data',
    default = True,
    help = 'Use bidirectional data streams, i.e., forward & backward.',
)
flags.DEFINE_integer(
    'mask_alpha', default = 6, help = 'How many tokens to form a group.'
)
flags.DEFINE_integer(
    'mask_beta',
    default = 1,
    help = 'How many tokens to mask within each group.',
)
flags.DEFINE_integer(
    'num_predict',
    default = None,
    help = 'Number of tokens to predict in partial prediction.',
)
flags.DEFINE_integer('perm_size', default = None, help = 'perm size.')
flags.DEFINE_bool('uncased', False, help = 'Use uncased inputs or not.')
flags.DEFINE_integer('n_token', 32000, help = 'Vocab size')

# Model config
flags.DEFINE_integer('mem_len', default = 0, help = 'Number of steps to cache')
flags.DEFINE_bool(
    'same_length', default = False, help = 'Same length attention'
)
flags.DEFINE_integer('clamp_len', default = -1, help = 'Clamp length')

flags.DEFINE_integer('n_layer', default = 6, help = 'Number of layers.')
flags.DEFINE_integer('d_model', default = 32, help = 'Dimension of the model.')
flags.DEFINE_integer(
    'd_embed', default = 32, help = 'Dimension of the embeddings.'
)
flags.DEFINE_integer('n_head', default = 4, help = 'Number of attention heads.')
flags.DEFINE_integer(
    'd_head', default = 8, help = 'Dimension of each attention head.'
)
flags.DEFINE_integer(
    'd_inner',
    default = 32,
    help = 'Dimension of inner hidden size in positionwise feed-forward.',
)
flags.DEFINE_float('dropout', default = 0.0, help = 'Dropout rate.')
flags.DEFINE_float('dropatt', default = 0.0, help = 'Attention dropout rate.')
flags.DEFINE_bool(
    'untie_r', default = False, help = 'Untie r_w_bias and r_r_bias'
)
flags.DEFINE_string(
    'summary_type',
    default = 'last',
    help = 'Method used to summarize a sequence into a compact vector.',
)
flags.DEFINE_string(
    'ff_activation',
    default = 'relu',
    help = 'Activation type used in position-wise feed-forward.',
)
flags.DEFINE_bool('use_bfloat16', False, help = 'Whether to use bfloat16.')

# Parameter initialization
flags.DEFINE_enum(
    'init',
    default = 'normal',
    enum_values = ['normal', 'uniform'],
    help = 'Initialization method.',
)
flags.DEFINE_float(
    'init_std', default = 0.02, help = 'Initialization std when init is normal.'
)
flags.DEFINE_float(
    'init_range',
    default = 0.1,
    help = 'Initialization std when init is uniform.',
)


FLAGS = flags.FLAGS


def per_device_batch_size(batch_size, num_gpus):
    """For multi-gpu, batch-size must be a multiple of the number of GPUs.
  Note that this should eventually be handled by DistributionStrategies
  directly. Multi-GPU support is currently experimental, however,
  so doing the work here until that feature is in place.
  Args:
    batch_size: Global batch size to be divided among devices. This should be
      equal to num_gpus times the single-GPU batch_size for multi-gpu training.
    num_gpus: How many GPUs are used with DistributionStrategies.
  Returns:
    Batch size per device.
  Raises:
    ValueError: if batch_size is not divisible by number of devices
  """
    if num_gpus <= 1:
        return batch_size

    remainder = batch_size % num_gpus
    if remainder:
        err = (
            'When running with multiple GPUs, batch size '
            'must be a multiple of the number of available GPUs. Found {} '
            'GPUs with a batch size of {}; try --batch_size={} instead.'
        ).format(num_gpus, batch_size, batch_size - remainder)
        raise ValueError(err)
    return int(batch_size / num_gpus)


def get_model_fn():
    """doc."""

    def model_fn(features, labels, mode, params):
        """doc."""
        #### Training or Evaluation
        is_eval = mode == tf.estimator.ModeKeys.EVAL
        assert is_eval

        #### Retrieve `mems` from `params["cache"]`
        mems = {}
        idx = 0
        if FLAGS.mem_len > 0:
            mems['mems'] = params['cache']

        #### Get loss from inputs
        total_loss, total_accuracy, new_mems, monitor_dict = custom_function_builder.get_loss(
            FLAGS, features, labels, mems, False
        )

        #### Turn `new_mems` into `new_cache`
        new_cache = []
        if FLAGS.mem_len > 0:
            new_cache += new_mems['mems']

        #### Check model parameters
        num_params = sum([np.prod(v.shape) for v in tf.trainable_variables()])
        tf.logging.info('#params: {}'.format(num_params))

        #### Configuring the optimizer
        train_op, learning_rate, gnorm = model_utils.get_train_op(
            FLAGS, total_loss
        )
        monitor_dict['lr'] = learning_rate
        monitor_dict['gnorm'] = gnorm

        #### Customized initial checkpoint
        scaffold_fn = model_utils.init_from_checkpoint(
            FLAGS, global_vars = True
        )

        # def metric_fn(accuracy):
        #     return
        #
        # eval_metrics = (metric_fn, [total_accuracy])

        output_spec = tf.estimator.EstimatorSpec(
            mode = mode,
            loss = total_loss,
            train_op = train_op,
            eval_metric_ops = {'accuracy': total_accuracy},
            scaffold = scaffold_fn,
        )

        return output_spec

    return model_fn


def get_input_fn(split, batch_size):
    """doc."""
    assert split == 'train'

    input_fn, record_info_dict = data_utils.get_input_fn(
        tfrecord_dir = FLAGS.record_info_dir,
        split = split,
        bsz_per_host = batch_size,
        seq_len = FLAGS.seq_len,
        reuse_len = FLAGS.reuse_len,
        bi_data = FLAGS.bi_data,
        num_hosts = 1,
        num_core_per_host = 1,
        perm_size = FLAGS.perm_size,
        mask_alpha = FLAGS.mask_alpha,
        mask_beta = FLAGS.mask_beta,
        uncased = FLAGS.uncased,
        num_passes = FLAGS.num_passes,
        use_bfloat16 = FLAGS.use_bfloat16,
        num_predict = FLAGS.num_predict,
    )

    return input_fn, record_info_dict


def get_cache_fn(mem_len, batch_size):
    """doc."""
    tf_float = tf.bfloat16 if FLAGS.use_bfloat16 else tf.float32

    def cache_fn():
        mems = []
        if FLAGS.mem_len > 0:
            for _ in range(FLAGS.n_layer):
                zeros = tf.zeros(
                    [mem_len, batch_size, FLAGS.d_model], dtype = tf_float
                )
                mems.append(zeros)

        return mems

    if mem_len > 0:
        return cache_fn
    else:
        return None


def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)

    # Get corpus info
    FLAGS.n_token = data_utils.VOCAB_SIZE
    tf.logging.info('n_token {}'.format(FLAGS.n_token))

    if not tf.gfile.Exists(FLAGS.model_dir):
        tf.gfile.MakeDirs(FLAGS.model_dir)

    bsz_per_core = FLAGS.train_batch_size

    train_input_fn, train_record_info_dict = get_input_fn('train', bsz_per_core)
    tf.logging.info(
        'num of batches {}'.format(train_record_info_dict['num_batch'])
    )
    train_cache_fn = get_cache_fn(FLAGS.mem_len, bsz_per_core)
    tf.logging.info(train_cache_fn)

    log_every_n_steps = 10
    run_config = RunConfig(
        log_step_count_steps = log_every_n_steps,
        model_dir = FLAGS.model_dir,
        save_checkpoints_steps = FLAGS.save_steps,
        save_summary_steps = None,
    )
    model_fn = get_model_fn()
    tf.logging.info('Use normal Estimator')
    estimator = Estimator(
        model_fn = model_fn,
        params = {'batch_size': bsz_per_core, 'cache': None},
        config = run_config,
    )

    tf.logging.info('***** Running evaluation *****')
    tf.logging.info('  Batch size = %d', FLAGS.train_batch_size)
    estimator.evaluate(input_fn = train_input_fn, steps = 100)


if __name__ == '__main__':
    tf.app.run()



# https://en.wikipedia.org/wiki/List_of_Malaysian_dishes#Malaysian_dishes
malay = {
    'ambuyat',
    'asam pedas',
    'ayam bakar',
    'ayam goreng',
    'ayam masak kicap',
    'ayam masak merah',
    'bubur asyura',
    'bubur kacang hijau',
    'bubur lambuk',
    'bubur pedas',
    'gulai',
    'ikan bakar',
    'ikan goreng',
    'kangkung belacan',
    'ketupat',
    'laksa',
    'lemang',
    'lontong',
    'mee bandung muar',
    'nasi ambeng',
    'nasi campur',
    'nasi dagang',
    'nasi goreng',
    'nasi goreng pattaya',
    'nasi kerabu',
    'nasi lemak',
    'nasi tumpang',
    'nasi ulam',
    'otak otak',
    'pulot tartal',
    'rendang',
    'rojak',
    'roti jala',
    'satay',
    'satay celup',
    'sup kambing',
    'ulam',
    'nasi ayam',
}

chinese = {
    'bak kut teh',
    'bakkwa',
    'banmian',
    'bean sprouts chicken',
    'bubur pulut hitam',
    'chai tow kway',
    'char kway teow',
    'char siu',
    'chee cheong fun',
    'chwee kueh',
    'claypot chicken rice',
    'duck soup noodles',
    'economy rice',
    'fish ball',
    'hae mee',
    'hainanese chicken rice',
    'hokkien mee',
    'kaya toast',
    'kolo mee',
    'loh bak',
    'lor mee',
    'mee pok',
    'mixed rice',
    'oyster omelette',
    'pan mee',
    'pao',
    'popiah',
    'tong sui',
    'wonton mee',
    'yong tau foo',
    'you char kway',
    'yusheng',
    'asam laksa',
    'bubur cha cha',
    'cap cai',
    'curry mee',
    'laksa lemak',
    'mee siam',
    'nyonya bak chang',
}

indian = {
    'adhirasam',
    'appam',
    'bajji',
    'banana leaf rice',
    'chapati',
    'curd rice',
    'dalcha',
    'dosai',
    'fish head curry',
    'fish molee',
    'idli',
    'korma',
    'maggi goreng',
    'mamak',
    'mee goreng',
    'mee goreng mamak',
    'mee rebus',
    'murtabak',
    'murukku',
    'naan',
    'nasi biryani',
    'nasi kandar',
    'paniyaram',
    'papadum',
    'pasembur',
    'payasam',
    'pongal',
    'puliyodarai',
    'puri',
    'putu mayam',
    'rasam',
    'rojak',
    'roti canai',
    'roti prata',
    'roti tissue',
    'sambar',
    'sup kambing',
    'upma',
    'vadai',
}

sabah_sarawak = {
    'bosou',
    'hinava',
    'linatan',
    'linongot',
    'manok pansoh',
    'midin',
    'nasi kombos',
    'nasi laru',
    'nasik aruk',
    'sinalau bakas',
    'tonokon',
    'tuhau',
    'umai',
}

kuih = {
    'ang ku',
    'bahulu',
    'kek batik',
    'heong peng',
    'kek sarawak',
    'amplang',
    'lekor',
    'mee siput muar',
    'rempeyek',
    'apam balik',
    'bingka',
    'borasa',
    'cakoi',
    'kuih cincin',
    'cucur',
    'dodol',
    'kuih gelang',
    'kuih jala',
    'kuih jelurut',
    'kuih jemput jemput',
    'karipap',
    'kuih kelupis',
    'kochi',
    'penyaram',
    'pisang goreng',
    'kuih gulung',
    'kuih kasturi',
    'kuih lapis',
    'kuih lidah',
    'kuih makmur',
    'ondeh ondeh',
    'otokon',
    'kuih pais',
    'pie tee',
    'pulut inti',
    'pulut panggang',
    'putu bambu',
    'putu bumbong',
    'putu mangkuk',
    'kuih gulong',
    'kuih kapit',
    'kuih sapit',
    'seri muka',
}

dessert = {'lamban', 'punjung', 'tapai', 'wajid'}

malaysian_food = malay | chinese | indian | sabah_sarawak | kuih | dessert

hot_ice_beverage = {
    'teh',
    'milo',
    'neslo',
    'kopi',
    'soya',
    'susu',
    'sirap',
    'kosong',
}

hot_ice_beverage_regex = '(?:\s*|ice|hot|cold|ais|panas|sejuk)\s*(?:susu|teh|kosong|soya|kopi|neslo|milo|sirap|coffee|tea)\s*(?:o|\s*)\s*(?:ice|sejuk|hot|panas|cold|ais|\s*)\s*(?:tarik|pull|)\\b'
fruit_juice_regex = '(?:juice|jus)\s*(?:durian|kelapa|manggis|strawberry|nanas|langsat|nangka|cempedak|kedondong|pisang|laici|betik|tembikai|jambu|ciku|nenas|mangga|orange|watermelon|rambutan|strawberri|apple|guava|epal|oren|mango|longan|asam|belimbing|carrot|limau)|(?:durian|kelapa|manggis|strawberry|nanas|langsat|nangka|cempedak|kedondong|pisang|laici|betik|tembikai|jambu|ciku|nenas|mangga|orange|watermelon|rambutan|strawberri|apple|guava|epal|oren|mango|longan|asam|belimbing|carrot|limau)\s*(?:juice|jus)\\b'


unique_beverage = {
    'ais kacang',
    'asam boi',
    'leng chee kang',
    'barley ais',
    'tau hua',
    'cendol',
    'bubur cha bha',
    'tong sui',
    'tealive',
    'bubble tea',
    'milk tea',
    'chatime',
    'boba',
    'laici kang',
}

unique_beverage_regex = '(?:' + '|'.join(list(unique_beverage)) + ')'

fruits = {
    'durian',
    'manggis',
    'longan',
    'laici',
    'rambutan',
    'langsat',
    'cempedak',
    'guava',
    'jambu',
    'betik',
    'belimbing',
    'tembikai',
    'watermelon',
    'nangka',
    'strawberry',
    'strawberri',
    'mangga',
    'mango',
    'nanas',
    'nenas',
    'ciku',
    'apple',
    'epal',
    'carrot',
    'pisang',
    'kelapa',
    'limau',
    'asam',
    'kedondong',
    'oren',
    'orange',
}

american = {
    'hotdog',
    'pie',
    'cookie',
    'cobbler',
    'nacho',
    'stew',
    'barbecue',
    'lime pie',
    'tots',
    'tater',
    'sourdough',
    'cobb',
    'salad',
    'pot roast',
    'twinkies',
    'jerky',
    'fajitas',
    'banana split',
    'cornbread',
    'jambalaya',
    'steak',
    'salmon',
    'meatloaf',
    'grit',
    'macaroni',
    'cheese',
    'crabcake',
    'chip',
    'cioppino',
    'bean',
    'popcorn',
    'ayam goreng',
    'fried chicken',
    'waffle',
}

italian = {
    'pizza',
    'focaccia',
    'carbonara',
    'pizza',
    'risotto',
    'pasta',
    'noodle',
    'tiramisu',
    'lasagne',
}

# https://en.wikipedia.org/wiki/Indonesian_cuisine

indonesian = {
    'acar',
    'ayam bakar',
    'ayam goreng',
    'ayam kecap',
    'bubur ayam',
    'bubur kacang hijau',
    'bubur ketan hitam',
    'gado gado',
    'ikan asin',
    'ikan bakar',
    'ikan goreng',
    'ketupat',
    'lontong',
    'nasi bakar',
    'nasi campur',
    'nasi goreng',
    'nasi kuning',
    'otak otak',
    'perkedel',
    'pindang',
    'rijsttafel',
    'rujak',
    'sambal',
    'sambal goreng udang',
    'sate',
    'semur',
    'sop buntut',
    'soto',
    'soto mie',
    'tahu goreng',
    'telur pindang',
    'tempeh goreng',
    'tempe goreng',
    'tumis kangkung',
}

acehnese = {'mie aceh', 'mie caluk', 'nasi gurih', 'roti canai', 'timphan'}

balinese = {'babi guling', 'betutu', 'lawar', 'nasi bali', 'sate lilit'}

bataknese = {
    'arsik',
    'babi panggang',
    'bentidali ni horbo',
    'dengke mas na niura',
    'itak gurgur',
    'lampet',
    'manuk napinadar',
    'na nidugu',
    'na tinombur',
    'ombusombus',
    'pohulpohul',
    'saksang',
    'sambal',
    'sasagun',
    'tanggotanggo',
    'tipatipa',
    'tuktuk',
}

betawi = {
    'asinan betawi',
    'kerak telor',
    'ketoprak',
    'ketupat sayur',
    'laksa betawi',
    'mie kangkung',
    'nasi uduk',
    'nasi ulam',
    'soto betawi',
}

bugis_makassar = {
    'burasa',
    'coto makassar',
    'gogos',
    'kapurung',
    'konro',
    'pallubasa',
    'sop saudara',
}

cirebonese = {
    'docang',
    'empal gentong',
    'mie koclok',
    'sega jamblang',
    'sega lengko',
    'tahu gejrot',
}

chinese_peranakan = {
    'babi kecap',
    'bakmi',
    'bakpau',
    'bakso',
    'cap cai',
    'fu yung hai',
    'i fu mie',
    'kepiting saus tiram',
    'kwetiau goreng',
    'laksa',
    'lontong cap go meh',
    'lumpia',
    'mie ayam',
    'mie goreng',
    'mie kering',
    'mun tahu',
    'nasi tim',
    'sapo tahu',
    'sekba',
    'siomay',
    'swikee',
}

javanese = {
    'ayam penyet',
    'botok',
    'buntil',
    'gudeg',
    'iga penyet',
    'krechek',
    'kuluban',
    'mie jawa',
    'mie rebus',
    'nasi ambeng',
    'nasi bogana',
    'nasi kucing',
    'nasi liwet',
    'nasi pecel',
    'opor',
    'pecel',
    'pecel lele',
    'rambak petis',
    'rawon',
    'sayur lodeh',
    'selat solo',
    'serundeng',
    'tahu campur',
    'tongseng',
    'trancam',
    'tumpeng',
    'urap',
}

minangkabau = {
    'asam pedas',
    'balado',
    'dendeng',
    'galamai',
    'gulai',
    'kepiting saus padang',
    'keripik sanjay',
    'lemang',
    'nasi kapau',
    'nasi padang',
    'rendang',
    'sate padang',
    'soto padang',
    'udang balado',
}

palembang = {
    'burgo',
    'laksan',
    'lakso',
    'mie celor',
    'nasi minyak',
    'pempek',
    'pindang',
    'tekwan',
    'tempoyak',
}

sasak = {'ayam taliwang', 'beberuk terong', 'plecing kangkung'}

sundanese = {
    'asinan bogor',
    'bandung',
    'batagor',
    'empal gepuk',
    'karedok',
    'kupat tahu',
    'laksa bogor',
    'lalab',
    'mie kocok',
    'nasi timbel',
    'oncom',
    'pepes',
    'sate bandeng',
    'sayur asem',
    'seblak',
    'soto',
    'tahu sumedang',
    'tauge goreng',
}

middle_eastern = {
    'hummus',
    'manakeesh',
    'halloumi',
    'meddamas',
    'falafel',
    'tabouleh',
    'moutabal',
    'ghanoush',
    'fattoush',
    'ummali',
    'shanklish',
    'shawarma',
    'tawook',
    'dolma',
    'kofta',
    'quwarmah',
    'mansaf',
    'kebab',
    'baklava',
    'knafeh',
    'masgouf',
}

indonesian_food = (
    indonesian
    | acehnese
    | balinese
    | bataknese
    | betawi
    | bugis_makassar
    | cirebonese
    | chinese_peranakan
    | javanese
    | minangkabau
    | palembang
    | sasak
    | sundanese
)

total_foods = (
    malaysian_food | indonesian_food | american | italian | middle_eastern
)

total_foods_regex = '(?:' + '|'.join(list(total_foods)) + ')'



import collections
import re
import numpy as np
import tensorflow as tf
from sklearn.utils import shuffle
from sklearn.manifold import TSNE
from scipy.spatial.distance import cdist
from tqdm import tqdm


def counter_words(sentences):
    word_counter = collections.Counter()
    word_list = []
    num_lines, num_words = (0, 0)
    for i in sentences:
        words = re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', i)
        word_counter.update(words)
        word_list.extend(words)
        num_lines += 1
        num_words += len(words)
    return word_counter, word_list, num_lines, num_words


def build_dict(word_counter, vocab_size = 50000):
    count = [['PAD', 0], ['UNK', 1], ['START', 2], ['END', 3]]
    count.extend(word_counter.most_common(vocab_size))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    return dictionary, {word: idx for idx, word in dictionary.items()}


def doc2num(word_list, dictionary):
    word_array = []
    unknown_val = len(dictionary)
    for word in word_list:
        word_array.append(dictionary.get(word, unknown_val))
    return np.array(word_array, dtype = np.int32)


def build_word_array(sentences, vocab_size):
    word_counter, word_list, num_lines, num_words = counter_words(sentences)
    dictionary, rev_dictionary = build_dict(word_counter, vocab_size)
    word_array = doc2num(word_list, dictionary)
    return word_array, dictionary, rev_dictionary, num_lines, num_words


def build_training_set(word_array):
    num_words = len(word_array)
    x = np.zeros((num_words - 4, 4), dtype = np.int32)
    y = np.zeros((num_words - 4, 1), dtype = np.int32)
    shift = np.array([-2, -1, 1, 2], dtype = np.int32)
    for idx in range(2, num_words - 2):
        y[idx - 2, 0] = word_array[idx]
        x[idx - 2, :] = word_array[idx + shift]
    return x, y


class Model:
    def __init__(self, graph_params):
        g_params = graph_params
        tf.reset_default_graph()
        self.sess = tf.InteractiveSession()
        self.X = tf.placeholder(tf.int64, shape = [None, 4])
        self.Y = tf.placeholder(tf.int64, shape = [None, 1])
        w_m2, w_m1, w_p1, w_p2 = tf.unstack(self.X, axis = 1)
        self.embed_weights = tf.Variable(
            tf.random_uniform(
                [g_params['vocab_size'], g_params['embed_size']],
                -g_params['embed_noise'],
                g_params['embed_noise'],
            )
        )
        embed_m2 = tf.nn.embedding_lookup(self.embed_weights, w_m2)
        embed_m1 = tf.nn.embedding_lookup(self.embed_weights, w_m1)
        embed_p1 = tf.nn.embedding_lookup(self.embed_weights, w_p1)
        embed_p2 = tf.nn.embedding_lookup(self.embed_weights, w_p2)
        embed_stack = tf.concat([embed_m2, embed_m1, embed_p1, embed_p2], 1)
        hid_weights = tf.Variable(
            tf.random_normal(
                [g_params['embed_size'] * 4, g_params['hid_size']],
                stddev = g_params['hid_noise']
                / (g_params['embed_size'] * 4) ** 0.5,
            )
        )
        hid_bias = tf.Variable(tf.zeros([g_params['hid_size']]))
        hid_out = tf.nn.tanh(tf.matmul(embed_stack, hid_weights) + hid_bias)
        self.nce_weights = tf.Variable(
            tf.random_normal(
                [g_params['vocab_size'], g_params['hid_size']],
                stddev = 1.0 / g_params['hid_size'] ** 0.5,
            )
        )
        nce_bias = tf.Variable(tf.zeros([g_params['vocab_size']]))
        self.cost = tf.reduce_mean(
            tf.nn.nce_loss(
                self.nce_weights,
                nce_bias,
                inputs = hid_out,
                labels = self.Y,
                num_sampled = g_params['neg_samples'],
                num_classes = g_params['vocab_size'],
                num_true = 1,
                remove_accidental_hits = True,
            )
        )
        self.logits = tf.argmax(
            tf.matmul(hid_out, self.nce_weights, transpose_b = True) + nce_bias,
            axis = 1,
        )
        if g_params['optimizer'] == 'RMSProp':
            self.optimizer = tf.train.RMSPropOptimizer(
                g_params['learn_rate']
            ).minimize(self.cost)
        elif g_params['optimizer'] == 'Momentum':
            self.optimizer = tf.train.MomentumOptimizer(
                g_params['learn_rate'], g_params['momentum']
            ).minimize(self.cost)
        elif g_params['optimizer'] == 'Adam':
            self.optimizer = tf.train.AdamOptimizer(
                g_params['learn_rate']
            ).minimize(self.cost)
        else:
            print('Optimizer not supported,exit.')
        self.sess.run(tf.global_variables_initializer())

    def train(self, X, Y, X_val, Y_val, epoch, batch_size):
        for i in range(epoch):
            X, Y = shuffle(X, Y)
            pbar = tqdm(
                range(0, len(X), batch_size), desc = 'train minibatch loop'
            )
            for batch in pbar:
                feed_dict = {
                    self.X: X[batch : min(batch + batch_size, len(X))],
                    self.Y: Y[batch : min(batch + batch_size, len(X))],
                }
                _, loss = self.sess.run(
                    [self.optimizer, self.cost], feed_dict = feed_dict
                )
                pbar.set_postfix(cost = loss)

            pbar = tqdm(
                range(0, len(X_val), batch_size), desc = 'test minibatch loop'
            )
            for batch in pbar:
                feed_dict = {
                    self.X: X_val[batch : min(batch + batch_size, len(X_val))],
                    self.Y: Y_val[batch : min(batch + batch_size, len(X_val))],
                }
                loss = self.sess.run(self.cost, feed_dict = feed_dict)
                pbar.set_postfix(cost = loss)
        return self.embed_weights.eval(), self.nce_weights.eval()


class Word2Vec:
    def __init__(self, embed_matrix, dictionary):
        self._embed_matrix = embed_matrix
        self._dictionary = dictionary
        self._reverse_dictionary = {v: k for k, v in dictionary.items()}

    def get_vector_by_name(self, word):
        return np.ravel(self._embed_matrix[self._dictionary[word], :])

    def n_closest(self, word, num_closest = 5, metric = 'cosine'):
        wv = self.get_vector_by_name(word)
        closest_indices = self.closest_row_indices(wv, num_closest + 1, metric)
        word_list = []
        for i in closest_indices:
            word_list.append(self._reverse_dictionary[i])
        if word in word_list:
            word_list.remove(word)
        return word_list

    def closest_row_indices(self, wv, num, metric):
        dist_array = np.ravel(
            cdist(self._embed_matrix, wv.reshape((1, -1)), metric = metric)
        )
        sorted_indices = np.argsort(dist_array)
        return sorted_indices[:num]

    def analogy(self, a, b, c, num = 1, metric = 'cosine'):
        va = self.get_vector_by_name(a)
        vb = self.get_vector_by_name(b)
        vc = self.get_vector_by_name(c)
        vd = vb - va + vc
        closest_indices = self.closest_row_indices(vd, num, metric)
        d_word_list = []
        for i in closest_indices:
            d_word_list.append(self._reverse_dictionary[i])
        return d_word_list

    def project_2d(self, start, end):
        tsne = TSNE(n_components = 2)
        embed_2d = tsne.fit_transform(self._embed_matrix[start:end, :])
        word_list = []
        for i in range(start, end):
            word_list.append(self._reverse_dictionary[i])
        return embed_2d, word_list




# coding: utf-8

# In[1]:


import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import numpy as np
from tqdm import tqdm
import re
import collections
import json
import os
from tensorflow.contrib import seq2seq


def sequence(s, w2v_model, maxlen = 50, vocabulary_size = 500000):
    words = s.split()
    np_array = np.zeros((maxlen), dtype = np.int32)
    current_no = 0
    for no, word in enumerate(words[: maxlen - 2]):
        id_to_append = 1
        if word in w2v_model:
            word_id = w2v_model[word]
            if word_id < vocabulary_size:
                id_to_append = word_id
        np_array[no] = id_to_append
        current_no = no
    np_array[current_no + 1] = 3
    return np_array


def batch_sequence(sentences, dictionary, maxlen = 50):
    np_array = np.zeros((len(sentences), maxlen), dtype = np.int32)
    for no_sentence, sentence in enumerate(sentences):
        current_no = 0
        for no, word in enumerate(sentence.split()[: maxlen - 2]):
            np_array[no_sentence, no] = dictionary.get(word, 1)
            current_no = no
        np_array[no_sentence, current_no + 1] = 3
    return np_array


class Model:
    def __init__(
        self,
        vocabulary_size,
        maxlen = 50,
        output_size = 512,
        learning_rate = 1e-3,
        embedding_size = 256,
        batch_size = 16,
        max_grad_norm = 10,
        **kwargs
    ):
        word_embeddings = tf.Variable(
            tf.random_uniform(
                [vocabulary_size, embedding_size], -np.sqrt(3), np.sqrt(3)
            )
        )
        self.output_size = output_size
        self.maxlen = maxlen
        self.embeddings = word_embeddings
        self.output_layer = tf.layers.Dense(vocabulary_size)
        self.output_layer.build(output_size)

        self.BEFORE = tf.placeholder(tf.int32, [None, maxlen])
        self.INPUT = tf.placeholder(tf.int32, [None, maxlen])
        self.AFTER = tf.placeholder(tf.int32, [None, maxlen])
        self.batch_size = tf.shape(self.INPUT)[0]

        self.get_thought = self.thought(self.INPUT)
        self.attention = tf.matmul(
            self.get_thought, tf.transpose(self.embeddings), name = 'attention'
        )
        self.fw_logits = self.decoder(self.get_thought, self.AFTER)
        self.bw_logits = self.decoder(self.get_thought, self.BEFORE)
        self.loss = self.calculate_loss(
            self.fw_logits, self.AFTER
        ) + self.calculate_loss(self.bw_logits, self.BEFORE)
        tvars = tf.trainable_variables()
        grads, _ = tf.clip_by_global_norm(
            tf.gradients(self.loss, tvars), max_grad_norm
        )
        self.optimizer = tf.train.AdamOptimizer(learning_rate).minimize(self.loss)

    def get_embedding(self, inputs):
        return tf.nn.embedding_lookup(self.embeddings, inputs)

    def thought(self, inputs):
        encoder_in = self.get_embedding(inputs)
        fw_cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        bw_cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        sequence_length = tf.reduce_sum(tf.sign(inputs), axis = 1)
        with tf.variable_scope(
            'thought_scope', reuse = False
        ):
            rnn_output = tf.nn.bidirectional_dynamic_rnn(
                fw_cell,
                bw_cell,
                encoder_in,
                sequence_length = sequence_length,
                dtype = tf.float32,
            )[1]
            return sum(rnn_output)

    def decoder(self, thought, labels):
        main = tf.strided_slice(labels, [0, 0], [self.batch_size, -1], [1, 1])
        shifted_labels = tf.concat([tf.fill([self.batch_size, 1], 2), main], 1)
        decoder_in = self.get_embedding(shifted_labels)
        cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        max_seq_lengths = tf.fill([self.batch_size], self.maxlen)
        helper = seq2seq.TrainingHelper(
            decoder_in, max_seq_lengths, time_major = False
        )
        decoder = seq2seq.BasicDecoder(cell, helper, thought)
        decoder_out = seq2seq.dynamic_decode(decoder)[0].rnn_output
        return decoder_out

    def calculate_loss(self, outputs, labels):
        mask = tf.cast(tf.sign(labels), tf.float32)
        logits = self.output_layer(outputs)
        return seq2seq.sequence_loss(logits, labels, mask)


def counter_words(sentences):
    word_counter = collections.Counter()
    word_list = []
    num_lines, num_words = (0, 0)
    for i in sentences:
        words = re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', i)
        word_counter.update(words)
        word_list.extend(words)
        num_lines += 1
        num_words += len(words)
    return word_counter, word_list, num_lines, num_words


def build_dict(word_counter, vocab_size = 500000):
    count = [['PAD', 0], ['UNK', 1], ['START', 2], ['END', 3]]
    count.extend(word_counter.most_common(vocab_size))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    return dictionary, {word: idx for idx, word in dictionary.items()}


def train_model(
    train_X,
    train_Y_before,
    train_Y_after,
    epoch = 10,
    batch_size = 16,
    embedding_size = 128,
    maxlen = 100,
    **kwargs
):
    word_counter, _, _, _ = counter_words(train_X)
    dictionary, _ = build_dict(word_counter)
    print(len(dictionary))
    _graph = tf.Graph()
    with _graph.as_default():
        model = Model(
            len(dictionary),
            embedding_size = embedding_size,
            output_size = embedding_size,
            batch_size = batch_size,
            maxlen = maxlen,
            **kwargs
        )
        sess = tf.InteractiveSession()
        saver = tf.train.Saver(tf.global_variables())
        sess.run(tf.global_variables_initializer())
    saver.save(sess, 'skip/model.ckpt')

    for e in range(epoch):
        pbar = tqdm(range(0, len(train_X), batch_size), desc = 'minibatch loop')
        for i in pbar:
            batch_x = batch_sequence(
                train_X[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y_before = batch_sequence(
                train_Y_before[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y_after = batch_sequence(
                train_Y_after[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            loss, _ = sess.run(
                [model.loss, model.optimizer],
                feed_dict = {
                    model.BEFORE: batch_y_before,
                    model.INPUT: batch_x,
                    model.AFTER: batch_y_after,
                },
            )
            pbar.set_postfix(cost = loss)
    saver.save(sess, 'skip/model.ckpt')
    return sess, model, dictionary


# In[2]:


import json
with open('news-bm.json','r') as fopen:
    corpus = json.loads(fopen.read())

print(len(corpus))
corpus = [sentence for sentence in corpus if len(sentence) > 10]
print(len(corpus))


# In[3]:


stride = 1
t_range = int((len(corpus) - 3) / stride + 1)
left, middle, right = [], [], []
for i in range(t_range):
    slices = corpus[i * stride : i * stride + 3]
    left.append(slices[0])
    middle.append(slices[1])
    right.append(slices[2])


# In[5]:


len(left) == len(middle) == len(right)


# In[6]:


from sklearn.utils import shuffle
left, middle, right = shuffle(left, middle, right)


# In[ ]:


_,_,dictionary = train_model(middle,left,right)
with open('skip-news-dict.json', 'w') as fopen:
    fopen.write(json.dumps(dictionary))